import time

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAssets(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/assets_and_image'

    def get_config_file(self):
        return 'test_asset_loading.yaml'

    def test_machine_wide_asset_loading(self):

        # test that the images asset class gets built correctly
        self.assertTrue(hasattr(self.mc, 'images'))
        self.assertTrue(self.mc.asset_manager._asset_classes)
        self.assertEqual(self.mc.asset_manager._asset_classes[0].path_string, 'images')

        # tests that assets are registered as expected with various conditions

        # /images folder
        self.assertIn('image1', self.mc.images)  # .gif
        self.assertIn('image2', self.mc.images)  # .jpg
        self.assertIn('image3', self.mc.images)  # .png

        # test subfolders listed in assets:images machine-wide config folders
        self.assertIn('image4', self.mc.images)  # /images/preload
        self.assertIn('image5', self.mc.images)  # /images/on_demand

        # test images from subfolder not listed in assets:images
        self.assertIn('image11', self.mc.images)  # /images/custom1

        # test subfolder under another subfolder listed in assets:images
        self.assertIn('image14', self.mc.images)  # /images/preload/subfolder

        # test images from the images: section that have names configured to be
        # different from their file names
        self.assertIn('image_12_new_name', self.mc.images)  # image12.png
        # custom1/image13.png
        self.assertIn('image_13_new_name', self.mc.images)

        # test that the images that were renamed were not also loaded based on
        # their original names
        self.assertNotIn('image12', self.mc.images)
        self.assertNotIn('image13', self.mc.images)

        # test that config dicts are merged and/or overwritten properly

        # test custom k/v pair from default config based on the folder the
        # asset was in
        self.assertEqual(self.mc.images['image4'].config['test_key'],
                         'test_value')
        self.assertEqual(self.mc.images['image14'].config['test_key'],
                         'test_value')

        # test custom k/v pair from asset entry in the images: section
        self.assertEqual(self.mc.images['image3'].config['test_key'],
                         'test_value_override3')

        # same as above, but test that it also works when the asset name is
        # different from the file name
        self.assertEqual(
                self.mc.images['image_12_new_name'].config['test_key'],
                'test_value_override12')

        # Test that mode assets were loaded properly
        self.assertIn('image6', self.mc.images)
        self.assertIn('image7', self.mc.images)
        self.assertIn('image8', self.mc.images)
        self.assertIn('image9', self.mc.images)
        self.assertIn('image10', self.mc.images)

        # Make sure all the assets are loaded. Wait if not
        while (self.mc.asset_manager.num_assets_to_load <
                   self.mc.asset_manager.num_assets_loaded):
            time.sleep(.1)
            self.advance_time()

        # Need to wait a bit since the loading was a separate thread
        self.advance_time(.1)

        # Make sure the ones that should have loaded on startup actually loaded
        self.assertTrue(self.mc.images['image1'].loaded)
        self.assertFalse(self.mc.images['image1'].loading)
        self.assertFalse(self.mc.images['image1'].unloading)

        self.assertTrue(self.mc.images['image2'].loaded)
        self.assertFalse(self.mc.images['image2'].loading)
        self.assertFalse(self.mc.images['image2'].unloading)

        self.assertTrue(self.mc.images['image3'].loaded)
        self.assertFalse(self.mc.images['image3'].loading)
        self.assertFalse(self.mc.images['image3'].unloading)

        self.assertTrue(self.mc.images['image8'].loaded)
        self.assertFalse(self.mc.images['image8'].loading)
        self.assertFalse(self.mc.images['image8'].unloading)

        self.assertTrue(self.mc.images['image2'].loaded)
        self.assertFalse(self.mc.images['image2'].loading)
        self.assertFalse(self.mc.images['image2'].unloading)

        self.assertTrue(self.mc.images['image4'].loaded)
        self.assertFalse(self.mc.images['image4'].loading)
        self.assertFalse(self.mc.images['image4'].unloading)

        self.assertTrue(self.mc.images['image7'].loaded)
        self.assertFalse(self.mc.images['image7'].loading)
        self.assertFalse(self.mc.images['image7'].unloading)

        self.assertTrue(self.mc.images['image11'].loaded)
        self.assertFalse(self.mc.images['image11'].loading)
        self.assertFalse(self.mc.images['image11'].unloading)

        self.assertTrue(self.mc.images['image_12_new_name'].loaded)
        self.assertFalse(self.mc.images['image_12_new_name'].loading)
        self.assertFalse(self.mc.images['image_12_new_name'].unloading)

        self.assertTrue(self.mc.images['image_13_new_name'].loaded)
        self.assertFalse(self.mc.images['image_13_new_name'].loading)
        self.assertFalse(self.mc.images['image_13_new_name'].unloading)

        # Make sure the ones that should not have loaded on startup didn't load
        self.assertFalse(self.mc.images['image5'].loaded)
        self.assertFalse(self.mc.images['image5'].loading)
        self.assertFalse(self.mc.images['image5'].unloading)

        self.assertFalse(self.mc.images['image9'].loaded)
        self.assertFalse(self.mc.images['image9'].loading)
        self.assertFalse(self.mc.images['image9'].unloading)

        self.assertFalse(self.mc.images['image10'].loaded)
        self.assertFalse(self.mc.images['image10'].loading)
        self.assertFalse(self.mc.images['image10'].unloading)

        self.assertFalse(self.mc.images['image6'].loaded)
        self.assertFalse(self.mc.images['image6'].loading)
        self.assertFalse(self.mc.images['image6'].unloading)

        # Start the mode and make sure those assets load
        self.mc.modes['mode1'].start()
        self.advance_time()

        # Give it a second to load. This file is tiny, so it shouldn't take
        # this long
        time.sleep(.001)
        self.advance_time(.1)
        for x in range(10):
            if not self.mc.images['image9'].loaded or not self.mc.images['image6'].loaded:
                time.sleep(.1)
                self.advance_time(.1)

        self.assertTrue(self.mc.images['image9'].loaded)
        self.assertFalse(self.mc.images['image9'].loading)
        self.assertFalse(self.mc.images['image9'].unloading)

        self.assertTrue(self.mc.images['image6'].loaded)
        self.assertFalse(self.mc.images['image6'].loading)
        self.assertFalse(self.mc.images['image6'].unloading)

        # test mode stop which should unload those assets
        self.mc.modes['mode1'].stop()

        for x in range(10):
            if self.mc.images['image9'].loaded:
                self.assertTrue(self.mc.images['image9'].unloading)
                self.advance_time(.1)

        self.assertFalse(self.mc.images['image9'].loaded)
        self.assertFalse(self.mc.images['image9'].loading)
        self.assertFalse(self.mc.images['image9'].unloading)

    def test_random_asset_group(self):
        # three assets, no weights

        # make sure the asset group was created
        self.assertIn('group1', self.mc.images)

        # make sure the randomness is working. To test this, we request the
        # asset 10,000 times and then count the results and assume that each
        # should be 3,333 +- 500 just to make sure the test never fails/
        res = list()
        for x in range(10000):
            res.append(self.mc.images['group1'].image)

        self.assertAlmostEqual(3333, res.count(self.mc.images['image1']),
                               delta=500)
        self.assertAlmostEqual(3333, res.count(self.mc.images['image2']),
                               delta=500)
        self.assertAlmostEqual(3333, res.count(self.mc.images['image3']),
                               delta=500)

    def test_random_asset_group_with_weighting(self):
        # three assets, third one has a weight of 2

        # make sure the asset group was created
        self.assertIn('group2', self.mc.images)

        # make sure the randomness is working. To test this, we request the
        # asset 10,000 times and then count the results and assume that each
        # should be 3,333 +- 500 just to make sure the test never fails/
        res = list()
        for x in range(10000):
            res.append(self.mc.images['group2'].image)

        self.assertAlmostEqual(2500, res.count(self.mc.images['image1']),
                               delta=500)
        self.assertAlmostEqual(2500, res.count(self.mc.images['image2']),
                               delta=500)
        self.assertAlmostEqual(5000, res.count(self.mc.images['image3']),
                               delta=500)

    def test_sequence_asset_group(self):
        # three assets, no weights

        self.assertIn('group3', self.mc.images)

        # Should always return in order, 1, 2, 3, 1, 2, 3...
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image3'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image3'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group3'].image, self.mc.images['image3'])

    def test_sequence_asset_group_with_count(self):
        # three assets, no weights

        self.assertIn('group4', self.mc.images)

        # Should always return in order, 1, 1, 1, 1, 2, 2, 3, 1, 1, 1, 1 ...
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image3'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image1'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image2'])
        self.assertIs(self.mc.images['group4'].image, self.mc.images['image3'])

    def test_random_force_next(self):
        # random, except it ensures the same one does not show up twice in a
        # row

        self.assertIn('group5', self.mc.images)

        # do it 10,000 times just to be sure. :)
        last = self.mc.images['group5'].image
        res = list()

        for x in range(10000):
            image = self.mc.images['group5'].image
            self.assertIsNot(last, image)
            last = image

            res.append(image)

        # Also check that the weights were right

        # BTW these weights are non-intuitive since the last asset is not
        # considered for the next round. e.g. image1 = 1, image2 = 5,
        # image3 = 1, so you'd think they would be 1400, 7200, 1400, but in
        # reality, 50% of the time, asset2 is not in contention, so really
        # asset2 has a 6-to-1 (84%) chance of being selected 66% of the time,
        # but a 0% chance of being selected 33% of the time, etc. So trust that
        # these numbers are right. :)
        self.assertAlmostEqual(2733, res.count(self.mc.images['image1']),
                               delta=500)
        self.assertAlmostEqual(4533, res.count(self.mc.images['image2']),
                               delta=500)
        self.assertAlmostEqual(2733, res.count(self.mc.images['image3']),
                               delta=500)

    def test_random_force_all(self):
        # random, except it ensures the same one does not show up twice before
        # they're all shown

        self.assertIn('group6', self.mc.images)

        for x in range(1000):
            this_set = set()
            this_set.add(self.mc.images['group6'].image)
            this_set.add(self.mc.images['group6'].image)
            this_set.add(self.mc.images['group6'].image)

            self.assertEqual(len(this_set), 3)



#!python
#cython: embedsignature=True, language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
cimport cpython.pycapsule as pycapsule
import logging

from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.inline cimport lerpU8, in_out_quad
from mpfmc.core.audio.notification_message cimport *
from mpfmc.core.audio.track cimport *


# Max and min 16-bit audio sample values (used in audio mixing functions)
DEF MAX_AUDIO_VALUE_S16 = ((1 << (16 - 1)) - 1)
DEF MIN_AUDIO_VALUE_S16 = -(1 << (16 - 1))


# ---------------------------------------------------------------------------
#    Track base class
# ---------------------------------------------------------------------------
cdef class Track:
    """
    Track base class
    """

    def __cinit__(self, *args, **kw):
        """C constructor"""
        self.state = NULL
        self.device_id = 0

    def __init__(self, object mc, object audio_callback_data, str name, int track_num, int buffer_size, float volume=1.0):
        """
        Constructor
        Args:
            mc: The media controller app
            audio_callback_data: The AudioCallbackData pointer wrapped in a PyCapsule object
            name: The track name
            track_num: The track number (corresponds to the SDL_Mixer channel number)
            buffer_size: The length of the track audio buffer in bytes
            volume: The track volume (0.0 to 1.0)
        """
        self.log = logging.getLogger("Track")
        self.mc = mc
        self._name = name
        self._number = track_num
        self._events_when_stopped = None
        self._events_when_played = None
        self._events_when_paused = None

        # Allocate memory for the track state (common among all track types)
        self.state = <TrackState*> PyMem_Malloc(sizeof(TrackState))
        self.state.mix_callback_function = NULL
        self.state.type_state = NULL
        self.state.number = track_num
        self.state.buffer = <Uint8 *>PyMem_Malloc(buffer_size)
        self.state.buffer_size = buffer_size
        self.state.ducking_control_points = g_array_sized_new(False, True, sizeof(guint8), CONTROL_POINTS_PER_BUFFER)
        self.log.debug("Allocated track audio buffer (%d bytes)", buffer_size)

        # The easiest way to pass a C pointer in a constructor is to wrap it in a PyCapsule
        # (see https://docs.python.org/3.4/c-api/capsule.html).  This basically wraps the
        # pointer in a Python object. It can be extracted using PyCapsule_GetPointer.
        self.state.callback_data = <AudioCallbackData*>pycapsule.PyCapsule_GetPointer(audio_callback_data, NULL)

        self.state.status = track_status_playing
        self.state.fade_steps = 0
        self.state.fade_steps_remaining = 0
        new_volume = <Uint8>min(max(volume * SDL_MIX_MAXVOLUME, 0), SDL_MIX_MAXVOLUME)
        self.state.volume = new_volume
        self.state.fade_volume_current = new_volume
        self.state.fade_volume_start = new_volume
        self.state.fade_volume_target = new_volume

        self.state.notification_messages = NULL

    def __dealloc__(self):
        """Destructor"""
        SDL_LockAudio()
        g_array_free(self.state.ducking_control_points, True)
        PyMem_Free(self.state)
        SDL_UnlockAudio()

    def __repr__(self):
        return '<Track.{}.{}>'.format(self.number, self.name)

    cdef TrackState *get_state(self):
        return self.state

    property name:
        def __get__(self):
            return self._name

    property volume:
        def __get__(self):
            return round(self.state.volume / SDL_MIX_MAXVOLUME, 2)

    @property
    def type(self):
        raise NotImplementedError('Must be overridden in derived class')

    @property
    def number(self):
        """Return the track number"""
        cdef int number = -1
        if self.state != NULL:
            SDL_LockAudio()
            number = self.state.number
            SDL_UnlockAudio()
        return number

    @property
    def events_when_stopped(self):
        """Return the list of events that are posted when the track is stopped"""
        return self._events_when_stopped

    @events_when_stopped.setter
    def events_when_stopped(self, events):
        """Sets the list of events that are posted when the track is stopped"""
        self._events_when_stopped = events

    @property
    def events_when_played(self):
        """Return the list of events that are posted when the track is played"""
        return self._events_when_played

    @events_when_played.setter
    def events_when_played(self, events):
        """Sets the list of events that are posted when the track is played"""
        self._events_when_played = events

    @property
    def events_when_paused(self):
        """Return the list of events that are posted when the track is paused"""
        return self._events_when_paused

    @events_when_paused.setter
    def events_when_paused(self, events):
        """Sets the list of events that are posted when the track is paused"""
        self._events_when_paused = events

    @property
    def supports_in_memory_sounds(self):
        """Return whether or not track supports in-memory sounds"""
        raise NotImplementedError('Must be overridden in derived class')

    @property
    def supports_streaming_sounds(self):
        """Return whether or not track supports streaming sounds"""
        raise NotImplementedError('Must be overridden in derived class')

    @property
    def fading(self):
        """Return whether or not the track is currently fading"""
        cdef bint fading = False
        if self.state != NULL:
            SDL_LockAudio()
            fading = self.state.fade_steps_remaining > 0
            SDL_UnlockAudio()
        return fading

    def clear_context(self, context):
        """Stop all sounds played from the specified context."""
        raise NotImplementedError('Must be overridden in derived class')

    def set_volume(self, float volume, float fade_seconds = 0.0):
        """Sets the current track volume with an optional fade time"""
        cdef Uint8 new_volume = <Uint8>min(max(volume * SDL_MIX_MAXVOLUME, 0), SDL_MIX_MAXVOLUME)
        SDL_LockAudio()

        # Fades require special logic
        if fade_seconds > 0:
            if self.state.status == track_status_stopping or self.state.status == track_status_pausing:
                # Fade is ignored if track is in the process of stopping or pausing
                self.state.volume = new_volume
            else:
                # If the track is currently in the middle of a fade, the existing fade
                # will be interrupted and a new fade will be calculated from the current
                # point of the existing fade
                if self.state.fade_steps_remaining > 0:
                    self.log.debug("set_volume - Interrupting an existing fade on this track so "
                                   "start a new fade")

                self.log.debug("set_volume - Applying %s second fade to new volume level", str(fade_seconds))

                # Calculate fade
                self.state.fade_steps = <Uint32>(fade_seconds * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
                self.state.fade_steps_remaining = self.state.fade_steps
                self.state.fade_volume_start = self.state.fade_volume_current
                self.state.fade_volume_target = new_volume
                self.state.volume = new_volume
        else:
            self.state.volume = new_volume
            self.state.fade_volume_current = new_volume
            self.state.fade_volume_start = new_volume
            self.state.fade_volume_target = new_volume

        SDL_UnlockAudio()

    def play(self, float fade_in_seconds = 0.0):
        """
        Starts playing the track so it can begin processing sounds. Function has no effect if
        the track is already playing.
        Args:
            fade_in_seconds: The number of seconds to fade in the track
        """
        self.log.debug("play - Begin sound processing on track")

        SDL_LockAudio()

        # Play is only supported when a track is paused, stopped, or is in the process of stopping
        if self.state.status == track_status_paused or \
                        self.state.status == track_status_stopped or \
                        self.state.status == track_status_stopping:
            if fade_in_seconds > 0:
                # Calculate fade data (steps and volume)
                self.log.debug("play - Applying %s second fade in", str(fade_in_seconds))
                self.state.fade_steps = <Uint32>(fade_in_seconds * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
                self.state.fade_steps_remaining = self.state.fade_steps
                self.state.fade_volume_start = self.state.fade_volume_current
                self.state.fade_volume_target = self.state.volume
            else:
                # No fade will occur, simply set volume
                self.state.fade_steps = 0
                self.state.fade_steps_remaining = 0
                self.state.fade_volume_current = self.state.volume
                self.state.fade_volume_start = self.state.volume
                self.state.fade_volume_target = self.state.volume

            self.state.status = track_status_playing

            # Trigger any events
            if self.events_when_played is not None:
                for event in self.events_when_played:
                    self.mc.post_mc_native_event(event, track=self._name)
        else:
            self.log.warning("play - Action may only be used when a track is stopped or is in the process "
                             "of stopping; action will be ignored.")

        SDL_UnlockAudio()

    def stop(self, float fade_out_seconds = 0.0):
        """
        Stops the track and clears out any playing sounds. Function has no effect if the track is
        already stopped.
        Args:
            fade_out_seconds: The number of seconds to fade out the track
        """
        self.log.debug("stop - Stop sound processing on track and clear state")

        SDL_LockAudio()

        # Stop is only supported when a track is playing
        if self.state.status in [track_status_playing, track_status_stopping, track_status_pausing]:

            if fade_out_seconds > 0:
                # Calculate fade data (steps and volume)
                self.log.debug("stop - Applying %s second fade out", str(fade_out_seconds))
                self.state.fade_steps = <Uint32>(fade_out_seconds * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
                self.state.fade_steps_remaining = self.state.fade_steps
                self.state.fade_volume_start = self.state.fade_volume_current
                self.state.fade_volume_target = 0
                self.state.status = track_status_stopping

            else:
                # No fade will occur, simply set volume
                self.state.fade_steps = 0
                self.state.fade_steps_remaining = 0
                self.state.fade_volume_current = 0
                self.state.fade_volume_start = 0
                self.state.fade_volume_target = 0
                self.state.status = track_status_stopped
                send_track_stopped_notification(self.state)

        else:
            self.log.warning("stop - Action may only be used when a track is playing; action "
                             "will be ignored.")

        SDL_UnlockAudio()

    def pause(self, float fade_out_seconds = 0.0):
        """
        Pauses the track. Sounds will continue from where they left off when the track is resumed.
        Function has no effect unless the track is playing.
        Args:
            fade_out_seconds: The number of seconds to fade out the track
        """
        self.log.debug("pause - Pause sound processing on track")

        SDL_LockAudio()

        # Stop is only supported when a track is playing
        if self.state.status in [track_status_playing, track_status_stopping, track_status_pausing]:
            if fade_out_seconds > 0:
                # Calculate fade data (steps and volume)
                self.log.debug("pause - Applying %s second fade out", str(fade_out_seconds))
                self.state.fade_steps = <Uint32>(fade_out_seconds * self.state.callback_data.seconds_to_bytes_factor) // self.state.callback_data.bytes_per_control_point
                self.state.fade_steps_remaining = self.state.fade_steps
                self.state.fade_volume_start = self.state.fade_volume_current
                self.state.fade_volume_target = 0
                self.state.status = track_status_pausing
            else:
                # No fade will occur, simply set volume
                self.state.fade_steps = 0
                self.state.fade_steps_remaining = 0
                self.state.fade_volume_current = 0
                self.state.fade_volume_start = 0
                self.state.fade_volume_target = 0
                self.state.status = track_status_paused
                send_track_paused_notification(self.state)

        else:
            self.log.warning("pause - Action may only be used when a track is playing; action "
                             "will be ignored.")

        SDL_UnlockAudio()

    def stop_all_sounds(self, float fade_out_seconds = 0.0):
        """
        Stops all playing sounds immediately on the track.
        Args:
            fade_out_seconds: The number of seconds to fade out the sounds before stopping
        """
        raise NotImplementedError('Must be overridden in derived class')

    def process(self):
        """Processes the track queue each tick."""
        raise NotImplementedError('Must be overridden in derived class')


    # ---------------------------------------------------------------------------
    #    Global C functions designed to be called from the static audio callback
    #    function (these functions do not use the GIL).
    #
    #    Note: Because these functions are only called from the audio callback
    #    function, we do not need to lock and unlock the mutex in these functions
    #    (locking/unlocking of the mutex is already performed in the audio
    #    callback function.
    # ---------------------------------------------------------------------------

    @staticmethod
    cdef void mix_track_to_output(TrackState *track, AudioCallbackData* callback_data,
                                  Uint8 *output_buffer, Uint32 buffer_length) nogil:
        """
        Applies ducking and mixes a track buffer into the master audio output buffer.
        Args:
            track: The track's state structure
            callback_data: The audio callback data structure
            output_buffer: The master audio output buffer.
            buffer_length: The audio buffer size to process.
    
        """

        cdef Uint8 *track_buffer
        cdef Uint32 output_buffer_bytes_remaining = buffer_length
        cdef Uint32 current_chunk_bytes
        cdef Uint32 buffer_pos = 0
        cdef Uint8 control_point = 0
        cdef Uint8 ducking_volume
        cdef Uint8 track_volume

        if track == NULL or track.status == track_status_stopped or track.status == track_status_paused:
            return

        track_buffer = <Uint8*>track.buffer

        # Loop over output buffer at control rate
        while output_buffer_bytes_remaining > 0:

            # Determine the number of bytes to process in the current chunk
            current_chunk_bytes = min(output_buffer_bytes_remaining, callback_data.bytes_per_control_point)

            # Apply any ducking to the track
            if track.ducking_is_active and control_point < CONTROL_POINTS_PER_BUFFER:
                ducking_volume = g_array_index_uint8(track.ducking_control_points, control_point)
            else:
                ducking_volume = SDL_MIX_MAXVOLUME

            # Calculate track volume (handle track fading)
            if track.fade_steps_remaining > 0:
                # Note: if the volume interpolation function below appears to be backwards, it is
                # because the fraction is going from 1 to 0 over the fade and not from a more
                # traditional 0 to 1.  This saves a few calculation cycles and is for efficiency.
                track.fade_volume_current = <Uint8> (lerpU8(in_out_quad(track.fade_steps_remaining / track.fade_steps),
                                                            track.fade_volume_target, track.fade_volume_start))
                track.fade_steps_remaining -= 1
            else:
                track.fade_volume_current = track.fade_volume_target
                if track.status == track_status_stopping:
                    track.status = track_status_stopped
                    send_track_stopped_notification(track)
                elif track.status == track_status_pausing:
                    track.status = track_status_paused
                    send_track_paused_notification(track)

            track_volume = ducking_volume * track.fade_volume_current // SDL_MIX_MAXVOLUME

            Track.mix_audio(output_buffer + buffer_pos, track_buffer + buffer_pos, current_chunk_bytes, track_volume)

            output_buffer_bytes_remaining -= current_chunk_bytes
            buffer_pos += current_chunk_bytes
            control_point += 1

    @staticmethod
    cdef void mix_audio(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length, int volume) nogil:
        """
        Mixes input buffer samples into existing output buffer samples applying the specified volume setting.
        
        Args:
            output_buffer: Output audio buffer
            input_buffer: Input audio buffer
            buffer_length: Length of audio buffers to process (input and output)
            volume: Volume level to apply to input buffer samples 

        """
        cdef Sample16 input_sample, output_sample
        cdef int temp_sample
        cdef Uint32 buffer_pos = 0

        while buffer_pos < buffer_length:

            # Get sample (2 bytes) from the input buffer and combine into 16-bit value
            input_sample.bytes.byte0 = input_buffer[buffer_pos]
            input_sample.bytes.byte1 = input_buffer[buffer_pos + 1]

            # Get sample (2 bytes) from the output buffer and combine into 16-bit value
            output_sample.bytes.byte0 = output_buffer[buffer_pos]
            output_sample.bytes.byte1 = output_buffer[buffer_pos + 1]

            # Calculate the new output sample (mix the input sample multiplied by the volume with the existing output
            # sample). The temporary output sample is a 32-bit value to avoid overflow.
            temp_sample = output_sample.value + ((input_sample.value * volume) // SDL_MIX_MAXVOLUME)

            # Clip the temp sample back to a 16-bit value (will cause distortion if samples
            # on channel are outside of the allowable 16-bit int range)
            if temp_sample > MAX_AUDIO_VALUE_S16:
                temp_sample = MAX_AUDIO_VALUE_S16
            elif temp_sample < MIN_AUDIO_VALUE_S16:
                temp_sample = MIN_AUDIO_VALUE_S16

            # Write the new output sample back to the output buffer (from
            # a 32-bit value back to a 16-bit value that we know is in 16-bit value range)
            output_sample.value = temp_sample
            output_buffer[buffer_pos] = output_sample.bytes.byte0
            output_buffer[buffer_pos + 1] = output_sample.bytes.byte1

            buffer_pos += 2

    @staticmethod
    cdef void mix_audio_stereo(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length,
                               int volume_left, int volume_right) nogil:
        """
        Mixes input buffer samples into existing output buffer samples applying the specified volume setting
        to the left and right channels.
        
        Args:
            output_buffer: Output audio buffer
            input_buffer: Input audio buffer
            buffer_length: Length of audio buffers to process (input and output)
            volume_left: Volume level to apply to the input buffer left channel samples 
            volume_right: Volume level to apply to the input buffer right channel samples
            
        Notes:
            SDL audio buffers are in interleaved format with the left channel first in stereo buffers.

        """
        cdef Sample16 input_sample, output_sample
        cdef int temp_sample
        cdef Uint32 buffer_pos = 0

        while buffer_pos < buffer_length:

            # Process left channel

            # Get sample (2 bytes) from the input buffer and combine into 16-bit value
            input_sample.bytes.byte0 = input_buffer[buffer_pos]
            input_sample.bytes.byte1 = input_buffer[buffer_pos + 1]

            # Get sample (2 bytes) from the output buffer and combine into 16-bit value
            output_sample.bytes.byte0 = output_buffer[buffer_pos]
            output_sample.bytes.byte1 = output_buffer[buffer_pos + 1]

            # Calculate the new output sample (mix the input sample multiplied by the volume with the existing output
            # sample). The temporary output sample is a 32-bit value to avoid overflow.
            temp_sample = output_sample.value + ((input_sample.value * volume_left) // SDL_MIX_MAXVOLUME)

            # Clip the temp sample back to a 16-bit value (will cause distortion if samples
            # on channel are outside of the allowable 16-bit int range)
            if temp_sample > MAX_AUDIO_VALUE_S16:
                temp_sample = MAX_AUDIO_VALUE_S16
            elif temp_sample < MIN_AUDIO_VALUE_S16:
                temp_sample = MIN_AUDIO_VALUE_S16

            # Write the new output sample back to the output buffer (from
            # a 32-bit value back to a 16-bit value that we know is in 16-bit value range)
            output_sample.value = temp_sample
            output_buffer[buffer_pos] = output_sample.bytes.byte0
            output_buffer[buffer_pos + 1] = output_sample.bytes.byte1

            buffer_pos += 2

            # Process right channel

            # Get sample (2 bytes) from the input buffer and combine into 16-bit value
            input_sample.bytes.byte0 = input_buffer[buffer_pos]
            input_sample.bytes.byte1 = input_buffer[buffer_pos + 1]

            # Get sample (2 bytes) from the output buffer and combine into 16-bit value
            output_sample.bytes.byte0 = output_buffer[buffer_pos]
            output_sample.bytes.byte1 = output_buffer[buffer_pos + 1]

            # Calculate the new output sample (mix the input sample multiplied by the volume with the existing output
            # sample). The temporary output sample is a 32-bit value to avoid overflow.
            temp_sample = output_sample.value + ((input_sample.value * volume_right) // SDL_MIX_MAXVOLUME)

            # Clip the temp sample back to a 16-bit value (will cause distortion if samples
            # on channel are outside of the allowable 16-bit int range)
            if temp_sample > MAX_AUDIO_VALUE_S16:
                temp_sample = MAX_AUDIO_VALUE_S16
            elif temp_sample < MIN_AUDIO_VALUE_S16:
                temp_sample = MIN_AUDIO_VALUE_S16

            # Write the new output sample back to the output buffer (from
            # a 32-bit value back to a 16-bit value that we know is in 16-bit value range)
            output_sample.value = temp_sample
            output_buffer[buffer_pos] = output_sample.bytes.byte0
            output_buffer[buffer_pos + 1] = output_sample.bytes.byte1

            buffer_pos += 2

    @staticmethod
    cdef void apply_volume(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length, int volume) nogil:
        """
        Applies a volume level to the input buffer samples and copies them to the output buffer.
        Args:
            output_buffer: Output audio buffer
            input_buffer: Input audio buffer
            buffer_length: Length of audio buffers to process (input and output)
            volume: Volume level to apply to input buffer samples 

        """
        cdef Sint16* output_samples = <Sint16*>output_buffer
        cdef Sint16* input_samples = <Sint16*>input_buffer
        cdef Uint32 buffer_pos = 0

        # We will be iterating over the audio buffers two bytes at a time (16-bit sample values) so
        # effectively the buffer is half the length since we are casting the buffer pointer to 16-bit
        buffer_length = buffer_length // 2

        while buffer_pos < buffer_length:
            output_samples[buffer_pos] = (input_samples[buffer_pos] * volume) // SDL_MIX_MAXVOLUME
            buffer_pos += 1

    @staticmethod
    cdef void apply_volume_stereo(Uint8* output_buffer, const Uint8* input_buffer, Uint32 buffer_length,
                                  int volume_left, int volume_right) nogil:
        """
        Applies a volume level to the left and right channel of the input buffer samples and copies them 
        to the output buffer.
        
        Args:
            output_buffer: Output audio buffer
            input_buffer: Input audio buffer
            buffer_length: Length of audio buffers to process (input and output)
            volume_left: Volume level to apply to input buffer left channel samples 
            volume_right: Volume level to apply to input buffer right channel samples

        Notes:
            SDL audio buffers are in interleaved format with the left channel first in stereo buffers.

        """
        cdef Sint16* output_samples = <Sint16*>output_buffer
        cdef Sint16* input_samples = <Sint16*>input_buffer
        cdef Uint32 buffer_pos = 0

        # We will be iterating over the audio buffers two bytes at a time (16-bit sample values) so
        # effectively the buffer is half the length since we are casting the buffer pointer to 16-bit
        buffer_length = buffer_length // 2

        while buffer_pos < buffer_length:
            output_samples[buffer_pos] = (input_samples[buffer_pos] * volume_left) // SDL_MIX_MAXVOLUME
            buffer_pos += 1

            output_samples[buffer_pos] = (input_samples[buffer_pos] * volume_right) // SDL_MIX_MAXVOLUME
            buffer_pos += 1



# pylint: disable=too-many-lines
"""Contains sound-related asset classes used by the audio system"""

import logging
import sys
import uuid
import weakref
from enum import Enum, unique
from typing import Optional, Union


from mpf.core.assets import AssetPool
from mpf.core.utility_functions import Util
from mpfmc.core.audio.audio_interface import AudioInterface

from mpfmc.assets.mc_asset import McAsset
from mpfmc.core.audio.audio_exception import AudioException


# ---------------------------------------------------------------------------
#    Default sound asset configuration parameter values
# ---------------------------------------------------------------------------
DEFAULT_VOLUME = 0.5
DEFAULT_PRIORITY = 0
DEFAULT_MAX_QUEUE_TIME = None
DEFAULT_LOOPS = 0
MINIMUM_DUCKING_DURATION = "10ms"


class SoundPool(AssetPool):
    """Allows several Sound assets to be grouped together in a 'pool' and referenced
    as a single sound when playing back.  Allows for easily managed variations
    of a sound.
    """
    # Be sure the pool group, if you use it, is first in the file ahead of the
    # asset class.
    # pylint: disable=invalid-name
    def __init__(self, mc, name, config, member_cls):
        super().__init__(mc, name, config, member_cls)

        self._track = None
        self._simultaneous_limit = None
        self._stealing_method = SoundStealingMethod.oldest
        self._key = None
        self.log = logging.getLogger('SoundPool')

        config.setdefault('track', None)
        config.setdefault('key', None)

        # Make sure a legal track name has been specified (unless only one track exists)
        if config['track']:
            # Track name specified in config, validate it
            track = self.machine.sound_system.audio_interface.get_track_by_name(self.config['track'])
            if track is None:
                self.log.error("'%s' is not a valid track name. "
                               "Could not create sound '%s' asset.", self.config['track'], name)
                raise AudioException("'{}' is not a valid track name. "
                                     "Could not create sound '{}' asset"
                                     .format(self.config['track'], name))
        else:
            # Track not specified, determine track count
            if self.machine.sound_system.audio_interface.get_track_count() == 1:
                # Only one track exists, assign default track
                track = self.machine.sound_system.audio_interface.get_track(0)
                config['track'] = track.name
                self.log.debug("Sound '%s' has been assigned a default track value of '%s' "
                               "(only one track has been configured)", name, track.name)
            else:
                raise AssertionError("SoundPool {} does not have a valid track.".format(name))

        self._track = config['track']
        self._key = config['key']

        if 'simultaneous_limit' in self.config and self.config['simultaneous_limit'] is not None:
            self._simultaneous_limit = int(self.config['simultaneous_limit'])

        if 'stealing_method' in self.config and self.config['stealing_method'] is not None:
            method = str(self.config['stealing_method']).lower()
            if method == 'skip':
                self._stealing_method = SoundStealingMethod.skip
            elif method == 'oldest':
                self._stealing_method = SoundStealingMethod.oldest
            elif method == 'newest':
                self._stealing_method = SoundStealingMethod.newest
            else:
                raise AudioException("Illegal value for sound_pool.stealing_method. "
                                     "Could not create sound pool '{}' asset".format(name))

    def __repr__(self):
        """String that's returned if someone prints this object"""
        return '<SoundPool: {} ({}), Loaded={}>'.format(self.name, self.id, self.loaded)

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        The id property contains a unique identifier for the sound pool (based on the Python id()
        function).  This id is used in the audio interface to uniquely identify a sound
        (rather than the name) due to the hassle of passing strings between Python and Cython.
        Return:
            An integer uniquely identifying the sound pool
        """
        return id(self)

    @property
    def loops(self):
        """Return loops."""
        return 0

    @property
    def key(self):
        """Return the unique key value for this sound pool"""
        return self._key

    @property
    def track(self):
        """The default track name on which to play the sound"""
        return self._track

    @property
    def sound(self):
        """The currently selected Sound object from the pool"""
        return self.asset

    @property
    def simultaneous_limit(self):
        """Return the maximum number of instances of the sound pool sounds that may be
        played simultaneously"""
        return self._simultaneous_limit

    @property
    def stealing_method(self):
        """Return the method used when stealing a sound instance (when a new sound
        instance is requested from the sound pool but the maximum number of instances
        has currently been reached)."""
        if self.simultaneous_limit is None:
            return None

        return self._stealing_method

    def play(self, context=None, settings=None):
        """
        Plays the sound pool on the track specified in the settings.
        Args:
            context: The context from which the sound is played.
            settings: A dictionary of settings for playback
        Returns:
            A SoundInstance object if the sound pool will be played (or queued for playback).
            None if the sound pool could not be played.
        """
        if settings and 'track' in settings:
            track_name = settings['track']
        else:
            track_name = self.track

        if track_name:
            track = self.machine.sound_system.audio_interface.get_track_by_name(track_name)
            if track:
                return track.play_sound(self, context, settings)

        return None

    def stop(self, fade_out=None):
        """
        Stops all instances of the the sound pool.

        Args:
            fade_out: Optional number of seconds over which to fade out before stopping. When
                      None, the SoundInstance value will be used.
        """
        if not fade_out:
            fade_out = 0.0

        self.machine.sound_system.audio_interface.stop_sound(self, fade_out)

    def stop_looping(self):
        """Stops all instances of the sound pool from continuing to loop."""
        self.machine.sound_system.audio_interface.stop_sound_looping(self)


@unique
class SoundStealingMethod(Enum):
    """Enumerated class containing sound stealing methods."""
    skip = 0    # Sound will be skipped (not played)
    oldest = 1  # The oldest sound will be replaced
    newest = 2  # The newest (most-recent) sound will be replaced


@unique
class ModeEndAction(Enum):
    """Enumerated class containing sound mode (context) end actions."""
    stop = 0          # Sound will stop immediately (uses fade_out setting)
    stop_looping = 1  # Looping will be canceled and the sound will be allowed to finish


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class SoundAsset(McAsset):
    """
    Sound asset class contains a single sound that may be played using the audio engine.

    Notes:
        It is critical that the AudioInterface be initialized before any Sound assets
        are loaded.  The loading code relies upon having an active audio interface.
    """
    attribute = 'sounds'  # attribute in MC, e.g. self.mc.images
    path_string = 'sounds'  # entry from mpf_mc:paths: for asset folder name
    config_section = 'sounds'  # section in the config files for this asset
    extensions = ('wav',)  # Additional extensions may be added at runtime
    class_priority = 100  # Order asset classes will be loaded. Higher is first.
    pool_config_section = 'sound_pools'  # Will setup groups if present
    asset_group_class = SoundPool  # Class or None to not use pools

    def __init__(self, mc, name, file, config):     # noqa
        """ Constructor"""
        super().__init__(mc, name, file, config)

        # Create a weakref dictionary to store all sound objects by id
        if not hasattr(self.machine, "sounds_by_id"):
            self.machine.sounds_by_id = weakref.WeakValueDictionary()
        self.machine.sounds_by_id[self.id] = self

        self._track = None
        self._streaming = False
        self._volume = DEFAULT_VOLUME
        self.priority = DEFAULT_PRIORITY
        self._max_queue_time = DEFAULT_MAX_QUEUE_TIME
        self._loops = DEFAULT_LOOPS
        self._loop_start_at = 0
        self._loop_end_at = None
        self._start_at = 0
        self._fade_in = 0
        self._fade_out = 0
        self._about_to_finish_time = None
        self._simultaneous_limit = None
        self._stealing_method = SoundStealingMethod.oldest
        self._events_when_played = None
        self._events_when_stopped = None
        self._events_when_looping = None
        self._events_when_about_to_finish = None
        self._mode_end_action = ModeEndAction.stop_looping
        self._markers = list()
        self._container = None  # holds the actual sound samples in memory
        self._ducking = None
        self._key = None
        self.log = logging.getLogger('SoundAsset')

        config.setdefault('track', None)

        # Make sure a legal track name has been specified (unless only one track exists)
        if config['track']:
            # Track name specified in config, validate it
            track = self.machine.sound_system.audio_interface.get_track_by_name(self.config['track'])
            if track is None:
                self.log.error("'%s' is not a valid track name. "
                               "Could not create sound '%s' asset.", self.config['track'], name)
                raise AudioException("'{}' is not a valid track name. "
                                     "Could not create sound '{}' asset"
                                     .format(self.config['track'], name))
        else:
            # Track not specified, determine track count
            if self.machine.sound_system.audio_interface.get_track_count() == 1:
                # Only one track exists, assign default track
                track = self.machine.sound_system.audio_interface.get_track(0)
                config['track'] = track.name
                self.log.debug("Sound '%s' has been assigned a default track value of '%s' "
                               "(only one track has been configured)", name, track.name)
            else:
                raise AssertionError("Sound {} does not have a valid track.".format(name))

        self._track = config['track']

        # Validate sound attributes and provide default values
        self.config.setdefault('streaming', False)
        self._streaming = self.config['streaming']

        self.config.setdefault('volume', 0.5)
        self._volume = min(max(float(self.config['volume']), 0.0), 1.0)

        self.config.setdefault('priority', 0)
        self.priority = int(self.config['priority'])

        self.config.setdefault('max_queue_time', None)
        if self.config['max_queue_time'] is None:
            self._max_queue_time = None
        else:
            self._max_queue_time = AudioInterface.string_to_secs(self.config['max_queue_time'])

        self.config.setdefault('loops', 0)
        self._loops = int(self.config['loops'])

        self.config.setdefault('loop_start_at', 0)
        self._loop_start_at = AudioInterface.string_to_secs(self.config['loop_start_at'])

        self.config.setdefault('loop_end_at', None)
        if self.config['loop_end_at'] is None:
            self._loop_end_at = None
        else:
            self._loop_end_at = AudioInterface.string_to_secs(self.config['loop_end_at'])

        self.config.setdefault('simultaneous_limit', None)
        if self.config['simultaneous_limit'] is None:
            self._simultaneous_limit = None
        else:
            self._simultaneous_limit = int(self.config['simultaneous_limit'])

        self.config.setdefault('start_at', 0)
        self._start_at = AudioInterface.string_to_secs(self.config['start_at'])

        self.config.setdefault('fade_in', 0)
        self._fade_in = AudioInterface.string_to_secs(self.config['fade_in'])

        self.config.setdefault('fade_out', 0)
        self._fade_out = AudioInterface.string_to_secs(self.config['fade_out'])

        self.config.setdefault('about_to_finish_time', None)
        if self.config['about_to_finish_time'] is None:
            self._about_to_finish_time = None
        else:
            self._about_to_finish_time = AudioInterface.string_to_secs(self.config['about_to_finish_time'])

        self.config.setdefault('stealing_method', 'oldest')
        method = str(self.config['stealing_method']).lower()
        if method == 'skip':
            self._stealing_method = SoundStealingMethod.skip
        elif method == 'oldest':
            self._stealing_method = SoundStealingMethod.oldest
        elif method == 'newest':
            self._stealing_method = SoundStealingMethod.newest
        else:
            raise AudioException("Illegal value for sound.stealing_method. "
                                 "Could not create sound '{}' asset".format(name))

        if 'events_when_played' in self.config and isinstance(
                self.config['events_when_played'], str):
            self._events_when_played = Util.string_to_event_list(self.config['events_when_played'])

        if 'events_when_stopped' in self.config and isinstance(
                self.config['events_when_stopped'], str):
            self._events_when_stopped = Util.string_to_event_list(self.config['events_when_stopped'])

        if 'events_when_looping' in self.config and isinstance(
                self.config['events_when_looping'], str):
            self._events_when_looping = Util.string_to_event_list(self.config['events_when_looping'])

        if 'events_when_about_to_finish' in self.config and isinstance(
                self.config['events_when_about_to_finish'], str):
            self._events_when_about_to_finish = Util.string_to_event_list(self.config['events_when_about_to_finish'])

        if 'mode_end_action' in self.config and self.config['mode_end_action'] is not None:
            action = str(self.config['mode_end_action']).lower()
            if action == 'stop':
                self._mode_end_action = ModeEndAction.stop
            elif action == 'stop_looping':
                self._mode_end_action = ModeEndAction.stop_looping
            else:
                raise AudioException("Illegal value for sound.mode_end_action. "
                                     "Could not create sound '{}' asset".format(name))

        if 'key' in self.config:
            self._key = self.config['key']

        if 'markers' in self.config:
            self._markers = SoundAsset.load_markers(self.config['markers'], self.name)

        if 'ducking' in self.config:
            try:
                self._ducking = DuckingSettings(self.machine, self.config['ducking'])
            except AudioException:
                raise AudioException("Error in ducking settings: {}. "
                                     "Could not create sound '{}' asset"
                                     .format(sys.exc_info()[1], self.name))

        elif self._track and "tracks" in self.machine.sound_system.config and \
                'ducking' in self.machine.sound_system.config['tracks'][self._track]:
            try:
                self._ducking = DuckingSettings(
                    self.machine, self.machine.sound_system.config['tracks'][self.config['track']]['ducking'])
            except AudioException:
                raise AudioException("Error in ducking settings: {} of track {}"
                                     "Could not create sound '{}' asset"
                                     .format(sys.exc_info()[1], self.config['track'], self.name))

        # An attenuation value of exactly 1.0 does absolutely nothing so
        # there is no point in keeping the ducking settings for this
        # sound when attenuation is 1.0.
        if self._ducking and self._ducking.attenuation == 1.0:
            self._ducking = None

    def __del__(self):
        """Destructor"""
        self.unload()

    def __repr__(self):
        """String that's returned if someone prints this object"""
        return '<Sound: {} ({}), Loaded={}>'.format(self.name, self.id, self.loaded)

    def __lt__(self, other):
        """Less than comparison operator"""
        # Note this is "backwards" (It's the __lt__ method but the formula uses
        # greater than because the PriorityQueue puts lowest first.)
        if other is None:
            return False

        return ("%s, %s" % (self.priority, self._id) >
                "%s, %s" % (other.priority, other.get_id()))

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        The id property contains a unique identifier for the sound (based on the Python id()
        function).  This id is used in the audio interface to uniquely identify a sound
        (rather than the name) due to the hassle of passing strings between Python and Cython.
        Return:
            An integer uniquely identifying the sound
        """
        return id(self)

    @property
    def streaming(self):
        """Return whether or not this sound streamed (if not it will be loaded in memory)"""
        return self._streaming

    @property
    def track(self):
        """The default track name on which to play the sound"""
        return self._track

    @property
    def volume(self):
        """Return the volume of the sound (float 0.0 to 1.0)"""
        return self._volume

    @property
    def start_at(self):
        """Return the start at time for the sound (in seconds)"""
        return self._start_at

    @property
    def loop_start_at(self):
        """Return the loop start at time for the sound (in seconds)."""
        return self._loop_start_at

    @property
    def loop_end_at(self):
        """
        Return the loop end at time for the sound (in seconds). If None, the loop end at point will be
        at the end of the sound.
        """
        return self._loop_end_at

    @property
    def fade_in(self):
        """Return the fade in time for the sound (in seconds)"""
        return self._fade_in

    @property
    def fade_out(self):
        """Return the fade out time for the sound (in seconds)"""
        return self._fade_out

    @property
    def about_to_finish_time(self):
        """
        Return the time before the end when an about to finish marker will be generated (in seconds).
        If set to None, no about to finish marker will be generated.
        """
        return self._about_to_finish_time

    @property
    def max_queue_time(self):
        """Return the maximum time a sound may be queued before
        playing or being discarded"""
        return self._max_queue_time

    @property
    def loops(self):
        """Return the looping setting for the sound.
        0 - do not loop, -1 loop infinitely, >= 1 the number of
        times to loop."""
        return self._loops

    @property
    def simultaneous_limit(self):
        """Return the maximum number of instances of the sound that may be
        played simultaneously"""
        if self.streaming:
            # Streamed sounds only support a single instance at a time, no matter what the
            # 'simultaneous_limit' setting is.
            return 1

        return self._simultaneous_limit

    @property
    def stealing_method(self):
        """Return the method used when stealing a sound instance (when a new sound
        instance is requested but the maximum number of instances has currently
        been reached)."""
        if self.simultaneous_limit is None:
            return None

        return self._stealing_method

    @property
    def events_when_played(self):
        """Return the list of events that are posted when the sound is played"""
        return self._events_when_played

    @property
    def events_when_stopped(self):
        """Return the list of events that are posted when the sound is stopped"""
        return self._events_when_stopped

    @property
    def events_when_looping(self):
        """Return the list of events that are posted when the sound begins a new loop"""
        return self._events_when_looping

    @property
    def events_when_about_to_finish(self):
        """Return the list of events that are posted when the sound is about to finish"""
        return self._events_when_about_to_finish

    @property
    def mode_end_action(self):
        """Return the action to take when the owning mode (context) ends"""
        return self._mode_end_action

    @property
    def stop_on_mode_end(self):
        """Return whether or not the sound should stop immediately upon mode end"""
        return self._mode_end_action == ModeEndAction.stop

    @property
    def markers(self):
        """List of marker dictionary objects containing markers for this sound (optional)"""
        return self._markers

    @property
    def marker_count(self):
        """Return the number of markers specified for the sound"""
        return len(self._markers)

    @property
    def container(self):
        """The container object wrapping the structure containing the actual sound data"""
        return self._container

    @property
    def ducking(self):
        """A DuckingSettings object containing the ducking settings for this sound (optional)"""
        return self._ducking

    @property
    def has_ducking(self):
        """Return whether or not this sound has ducking"""
        return self._ducking is not None

    @property
    def key(self):
        """Return the unique key value for this sound"""
        return self._key

    def create_instance(self, context: Optional[str] = None, settings: Optional[dict] = None) -> "SoundInstance":
        """Creates a new SoundInstance."""
        return SoundInstance(self, context, settings)

    def play(self, context=None, settings=None):
        """
        Plays the sound on the track specified in the settings.
        Args:
            context: The context from which the sound is played.
            settings: A dictionary of settings for playback
        Returns:
            A SoundInstance object if the sound will be played (or queued for playback).
            None if the sound could not be played.
        """
        if settings and 'track' in settings:
            track_name = settings['track']
        else:
            track_name = self.track

        if track_name:
            track = self.machine.sound_system.audio_interface.get_track_by_name(track_name)
            if track:
                return track.play_sound(self, context, settings)

        return None

    def stop(self, fade_out=None):
        """
        Stops all instances of the the sound.

        Args:
            fade_out: Optional number of seconds over which to fade out before stopping. When
                      None, the SoundAsset value will be used.
        """
        if not fade_out:
            fade_out = self.fade_out

        self.machine.sound_system.audio_interface.stop_sound(self, fade_out)

    def stop_looping(self):
        """Stops all instances of the sound from continuing to loop."""
        self.machine.sound_system.audio_interface.stop_sound_looping(self)

    def do_load(self):
        """Loads the sound asset from disk."""

        # Load the sound file into memory
        if self._container is not None:
            self.log.debug("Sound %s already loaded", self.name)
            return

        try:
            if self.streaming:
                self.log.debug("Sound %s loading for streaming", self.name)
                self._container = self.machine.sound_system.audio_interface.load_sound_file_for_streaming(self.file)
            else:
                self.log.debug("Sound %s loading to memory", self.name)
                self._container = self.machine.sound_system.audio_interface.load_sound_file_to_memory(self.file)

        except AudioException as exception:
            self.log.error("Load sound %s failed due to an exception - %s",
                           self.name, str(exception))
            raise AudioException("Load sound {} failed due to an exception: {}".format(
                self.name, exception)) from exception

        # Validate ducking now that the sound has been loaded
        # TODO: Implement me
        # if self._ducking is not None:
        #    try:
        #        self._ducking.validate(self._container.length)
        #    except AudioException as exception:
        #        self.log.error("Ducking settings for sound %s are not valid: %s",
        #                       self.name, str(exception))
        #        raise
        #

    def _do_unload(self):
        """Unloads the asset from memory"""
        self.log.debug("Sound %s unloading", self.name)
        self.stop(0)
        if self._container is not None:
            self.machine.sound_system.audio_interface.unload_sound_file(self._container)
            self._container = None

    def is_loaded(self):
        """Called when the asset has finished loading"""
        super().is_loaded()
        self.log.debug("Loaded %s (Track %s)", self.name, self.track)

    @staticmethod
    def load_markers(config, sound_name):
        """
        Load and validate the markers config section
        Args:
            config: The 'markers' configuration file section for the sound
            sound_name: The name of the sound

        Returns:
            List of sound marker dictionary objects
        """

        markers = list()

        if isinstance(config, dict):
            config_markers = list(config)
        elif isinstance(config, list):
            config_markers = config
        else:
            raise AudioException("Sound {} has an invalid markers section".format(sound_name))

        last_marker_time = 0

        # Loop over all markers in the list
        for settings in config_markers:
            marker = dict()

            # Set marker parameters
            marker['time'] = AudioInterface.string_to_secs(settings['time'])
            if marker['time'] < last_marker_time:
                raise AudioException("Sound markers for sound {} must be in ascending time order".format(sound_name))
            last_marker_time = marker['time']

            if 'events' in settings and settings['events'] is not None:
                marker['events'] = Util.string_to_event_list(settings['events'])
            else:
                raise AudioException("Sound markers for sound {} must specify at least one event".format(sound_name))

            if 'name' in settings and settings['name'] is not None:
                marker['name'] = settings['name'].lower()
            else:
                marker['name'] = None

            if len(markers) == AudioInterface.get_max_markers():
                raise AudioException("Cannot add marker - the limit of {} sound markers has been "
                                     "reached for sound {}.".format(AudioInterface.get_max_markers(), sound_name))

            markers.append(marker)

        return markers


@unique
class SoundInstanceStatus(Enum):
    """Enumerated class containing status values for SoundInstance class."""
    pending = 0
    queued = 1
    playing = 2
    finished = 3


# pylint: disable=too-many-public-methods
class SoundInstance:
    """An instance of a playing sound asset. This class is essentially a wrapper container
    for sound assets that contains all the overridden parameter values for playback."""

    def __init__(self, sound: Union[SoundAsset, SoundPool],     # noqa
                 context: Optional[str] = None, settings: Optional[dict] = None):
        """Construct sound instance."""
        if sound is None:
            raise ValueError("Cannot create sound instance: sound parameter is None")

        # pylint: disable=invalid-name
        self.mc = sound.machine
        self._timestamp = self.mc.clock.get_time()

        if context and isinstance(context, str):
            self._context = context
        else:
            self._context = None

        self._status = SoundInstanceStatus.pending
        self._pan = 0
        self._played = False
        self._loop_count = 0
        self._registered_finished_handlers = list()
        self.log = logging.getLogger('SoundInstance')
        self._sound_id = sound.id

        # The underlying sound must be accessed differently for SoundAsset objects
        # (just use the object itself) and SoundPool object (retrieve the next
        # SoundAsset object in the pool).
        if isinstance(sound, SoundAsset):
            self._sound = sound
        elif isinstance(sound, SoundPool):
            self._sound = sound.sound

        # Simultaneous limit comes from the SoundAsset or SoundPool class and may not
        # be overridden
        self._simultaneous_limit = sound.simultaneous_limit
        self._stealing_method = sound.stealing_method
        self._markers = self._sound.markers
        self._exp_time = None

        # Assign default values from parent sound for parameters that can be overridden
        self._track = sound.track
        self._loops = self._sound.loops
        self._volume = self._sound.volume
        self._priority = self._sound.priority
        self._start_at = self._sound.start_at
        self._loop_start_at = self._sound.loop_start_at
        self._loop_end_at = self._sound.loop_end_at
        self._fade_in = self._sound.fade_in
        self._fade_out = self._sound.fade_out
        self._about_to_finish_time = self._sound.about_to_finish_time
        self._max_queue_time = self._sound.max_queue_time
        self._events_when_played = self._sound.events_when_played
        self._events_when_stopped = self._sound.events_when_stopped
        self._events_when_looping = self._sound.events_when_looping
        self._events_when_about_to_finish = self._sound.events_when_about_to_finish
        self._mode_end_action = self._sound.mode_end_action
        self._key = sound.key

        if settings is None:
            settings = dict()

        # TODO: Implement parameter validation for overridden parameters

        # Assign any overridden parameter values
        if settings.get('track'):
            self._track = settings['track']

        if settings.get('loops'):
            self._loops = settings['loops']

        if settings.get('volume'):
            self._volume = settings['volume']

        if settings.get('pan'):
            self._pan = settings['pan']

        if settings.get('priority'):
            self._priority = settings['priority']

        if settings.get('start_at'):
            self._start_at = settings['start_at']

        if settings.get('loop_start_at'):
            self._loop_start_at = settings['loop_start_at']

        if settings.get('loop_end_at'):
            self._loop_end_at = settings['loop_end_at']

        if settings.get('fade_in'):
            self._fade_in = settings['fade_in']

        if settings.get('fade_out'):
            self._fade_out = settings['fade_out']

        if settings.get('about_to_finish_time', -1) != -1:
            self._about_to_finish_time = settings['about_to_finish_time']

        if settings.get('max_queue_time', -1) != -1:
            self._max_queue_time = settings['max_queue_time']

        if settings.get('events_when_played', ['use_sound_setting']) != ['use_sound_setting']:
            self._events_when_played = settings['events_when_played']

        if settings.get('events_when_stopped', ['use_sound_setting']) != ['use_sound_setting']:
            self._events_when_stopped = settings['events_when_stopped']

        if settings.get('events_when_looping', ['use_sound_setting']) != ['use_sound_setting']:
            self._events_when_looping = settings['events_when_looping']

        if settings.get('events_when_about_to_finish', ['use_sound_setting']) != ['use_sound_setting']:
            self._events_when_about_to_finish = settings['events_when_about_to_finish']

        if settings.get('mode_end_action', 'use_sound_setting') != 'use_sound_setting':
            action = str(settings['mode_end_action']).lower()
            if action == 'stop':
                self._mode_end_action = ModeEndAction.stop
            elif action == 'stop_looping':
                self._mode_end_action = ModeEndAction.stop_looping

        if settings.get('key', 'use_sound_setting') != 'use_sound_setting':
            self._key = settings['key']

    def __repr__(self):
        """String that's returned if someone prints this object"""
        return '<SoundInstance: {} ({}), Volume={}, Loops={}, Priority={}, Loaded={}, Track={}>'.format(
            self.sound.name, self.id, self.volume, self.loops, self.priority, self.sound.loaded, self.track)

    def __lt__(self, other):
        """Less than comparison operator"""
        # Note this is "backwards" (It's the __lt__ method but the formula uses
        # greater than because the PriorityQueue puts lowest first.)
        if other is None:
            return False
        if self.priority != other.priority:
            return self.priority > other.priority
        if self.exp_time is None and other.exp_time is not None:
            return True
        if self.exp_time is not None and other.exp_time is None:
            return False
        if self.exp_time != other.exp_time:
            return self.exp_time > other.exp_time
        return self.id > other.id

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        The id property contains a unique identifier for the sound reference(based on the Python
        id() function).  This id is used in the audio interface to uniquely identify a sound
        instance (rather than the name) due to the hassle of passing strings between Python and
        Cython.
        Return:
            An integer uniquely identifying the sound reference
        """
        return id(self)

    @property
    def timestamp(self):
        """Return the time at which the sound instance was created"""
        return self._timestamp

    @property
    def context(self):
        """Return the context in which the sound instance was created"""
        return self._context

    @property
    def exp_time(self):
        """Return the expiration time of the sound instance"""
        return self._exp_time

    @exp_time.setter
    def exp_time(self, value):
        """Set the expiration time of the sound instance"""
        self._exp_time = value

    @property
    def sound(self):
        """The sound asset wrapped by this object"""
        return self._sound

    @property
    def sound_id(self):
        """The parent sound asset (or sound pool) object id this instance was created from"""
        return self._sound_id

    @property
    def streaming(self):
        """Return whether or not the wrapped sound is streamed
        (if not it will be loaded in memory)"""
        return self._sound.streaming

    @property
    def loaded(self):
        """Return whether or not the underlying sound asset file is loaded into memory"""
        return self._sound.loaded

    @property
    def name(self):
        """The name of the sound"""
        return self._sound.name

    @property
    def track(self):
        """The audio track name on which to play the sound"""
        return self._track

    @property
    def volume(self):
        """Return the volume of the sound (float 0.0 to 1.0)"""
        return self._volume

    @property
    def pan(self):
        """Return the pan setting of the sound (float -1.0 to 1.0)"""
        return self._pan

    @property
    def priority(self):
        """Return the priority of the sound"""
        return self._priority

    @property
    def max_queue_time(self):
        """Return the maximum time a sound may be queued before
        playing or being discarded"""
        return self._max_queue_time

    @property
    def start_at(self):
        """Return the start at time for the sound (in seconds)"""
        return self._start_at

    @property
    def loop_start_at(self):
        """Return the loop start at time for the sound (in seconds)."""
        return self._loop_start_at

    @property
    def loop_end_at(self):
        """
        Return the loop end at time for the sound (in seconds). If None, the loop end at point will be
        at the end of the sound.
        """
        return self._loop_end_at

    @property
    def fade_in(self):
        """Return the fade in time for the sound (in seconds)"""
        return self._fade_in

    @property
    def fade_out(self):
        """Return the fade out time for the sound (in seconds)"""
        return self._fade_out

    @property
    def about_to_finish_time(self):
        """
        Return the time before the end when an about to finish marker will be generated (in seconds).
        If set to None, no about to finish marker will be generated.
        """
        return self._about_to_finish_time

    @property
    def loops(self):
        """Return the looping setting for the sound.
        0 - do not loop, -1 loop infinitely, >= 1 the number of
        times to loop."""
        return self._loops

    @property
    def simultaneous_limit(self):
        """Return the maximum number of instances of the sound that may be
        played simultaneously"""
        return self._simultaneous_limit

    @property
    def stealing_method(self):
        """Return the method used when stealing a sound instance (when a new sound
        instance is requested but the maximum number of instances has currently
        been reached)."""
        return self._stealing_method

    @property
    def events_when_played(self):
        """Return the list of events that are posted when the sound is played"""
        return self._events_when_played

    @property
    def events_when_stopped(self):
        """Return the list of events that are posted when the sound is stopped"""
        return self._events_when_stopped

    @property
    def events_when_looping(self):
        """Return the list of events that are posted when the sound begins a new loop"""
        return self._events_when_looping

    @property
    def events_when_about_to_finish(self):
        """Return the list of events that are posted when the sound is about to finish"""
        return self._events_when_about_to_finish

    @property
    def mode_end_action(self):
        """Return the action to take when the owning mode (context) ends"""
        return self._mode_end_action

    @property
    def stop_on_mode_end(self):
        """Return whether or not the sound should stop immediately upon mode end"""
        return self._mode_end_action == ModeEndAction.stop

    @property
    def markers(self):
        """Return the list of marker dictionary objects for the sound"""
        return self._markers

    @property
    def marker_count(self):
        """Return the number of markers specified for the sound"""
        return len(self._markers)

    @property
    def container(self):
        """The container object wrapping the SDL structure containing the actual sound data"""
        return self._sound.container

    @property
    def ducking(self):
        """A DuckingSettings object containing the ducking settings for this sound (optional)"""
        return self._sound.ducking

    @property
    def has_ducking(self):
        """Return whether or not this sound has ducking"""
        return self._sound.ducking is not None

    @property
    def key(self):
        """Return the unique key value for this sound instance"""
        return self._key

    @property
    def queued(self):
        """Indicates whether or not this sound reference is currently queued for playback."""
        return self._status == SoundInstanceStatus.queued

    @property
    def pending(self):
        """Indicates whether or not this sound instance is currently pending playback."""
        return self._status == SoundInstanceStatus.pending or \
            self._status == SoundInstanceStatus.queued

    @property
    def playing(self):
        """Return whether or not this sound instance is currently playing."""
        return self._status == SoundInstanceStatus.playing

    def stop(self, fade_out=None):
        """
        Stops the sound instance.

        Args:
            fade_out: Optional number of seconds over which to fade out before stopping. When
                      None, the SoundInstance value will be used.
        """
        if not fade_out:
            fade_out = self.fade_out

        if self.track is not None:
            track = self.mc.sound_system.audio_interface.get_track_by_name(self.track)
            if track:
                track.stop_sound_instance(self, fade_out)
        else:
            self.mc.sound_system.audio_interface.stop_sound_instance(self, fade_out)

    def stop_looping(self):
        """Stops the sound instance from continuing to loop."""
        if self.track is not None:
            track = self.mc.sound_system.audio_interface.get_track_by_name(self.track)
            if track:
                track.stop_sound_instance_looping(self)
        else:
            self.mc.sound_system.audio_interface.stop_sound_instance_looping(self)

    def set_pending(self):
        """Set the sound instance status to pending."""
        self._status = SoundInstanceStatus.pending

    def set_queued(self):
        """Notifies the sound instance that it is now queued and triggers any
        corresponding actions"""
        self._status = SoundInstanceStatus.queued

    def set_playing(self):
        """Notifies the sound instance that it is now playing and triggers any
        corresponding actions."""
        self._status = SoundInstanceStatus.playing
        self._played = True
        if self.events_when_played is not None:
            for event in self.events_when_played:
                self.mc.post_mc_native_event(event, sound_instance=self)

    def set_stopped(self):
        """Notifies the sound instance that it has now stopped and triggers any
        corresponding actions."""

        # Trigger any events
        if self.events_when_stopped is not None:
            for event in self.events_when_stopped:
                self.mc.post_mc_native_event(event, sound_instance=self)

        self._finished()

    def set_looping(self):
        """Notifies the sound instance that it is now looping and triggers any
        corresponding actions."""
        # Increment the total loop count (how many times has instance looped)
        self._loop_count += 1

        # Trigger any events
        if self.events_when_looping is not None:
            for event in self.events_when_looping:
                self.mc.post_mc_native_event(event, sound_instance=self)

    def set_about_to_finish(self):
        """Notifies the sound instance that it is about to finish and triggers any
        corresponding actions."""

        # Trigger any events
        if self.events_when_about_to_finish is not None:
            for event in self.events_when_about_to_finish:
                self.mc.post_mc_native_event(event, sound_instance=self)

    def set_marker(self, marker_id):
        """Notifies the sound instance that the specified marker has just been reached
        during playback and triggers any corresponding actions."""
        try:
            marker = self._markers[marker_id]
        except ValueError:
            self.log.warning("An invalid marker ID was received: %d", marker_id)
            return

        # Trigger any events
        if marker['events'] is not None:
            for event in marker['events']:
                self.mc.post_mc_native_event(event, sound_instance=self, marker_id=marker_id)

    def set_expired(self):
        """Notifies the sound instance that it has expired and will not be played."""
        self._finished()

    def set_canceled(self):
        """Notifies the sound instance that is has been canceled and will not be played."""
        self._finished()

    def add_finished_handler(self, handler, priority=1, **kwargs):
        """
        Register a handler to be called when the sound instance has finished.
        Args:
            handler: The method that will be called when the sound instance has finished
            priority: An arbitrary integer value that defines what order the handlers will
                be called in (highest to lowest).
            **kwargs: Any additional keyword/arguments pairs

        Returns:
            A GUID reference to the handler which you can use to later remove the handler
            via ``remove_finished_handler_by_key``.
        """
        if not callable(handler):
            raise ValueError("Cannot add finished handler '{}' for sound '{}'"
                             .format(handler, self.name))

        key = uuid.uuid4()

        # A 'handler' in our case is a tuple with 4 elements:
        # the handler method, priority, dict of kwargs, & uuid key

        self._registered_finished_handlers.append((handler, priority, kwargs, key))
        try:
            self.log.debug("Registered %s as a finished handler for sound '%s', priority: %s, "
                           "kwargs: %s",
                           (str(handler).split(' '))[2], self.name, priority, kwargs)
        except IndexError:
            pass

        # Sort the handlers for this event based on priority. We do it now
        # so the list is pre-sorted so we don't have to do that with each
        # event post.
        self._registered_finished_handlers.sort(key=lambda x: x[1], reverse=True)

        return key

    def remove_finished_handler_by_key(self, key):
        """Remove a registered finished handler by key.

        Args:
            key: The key of the handler you want to remove
        """
        for handler in self._registered_finished_handlers:
            if handler[3] == key:
                self._registered_finished_handlers.remove(handler)
                self.log.debug("Removing finished handler method %s from sound %s",
                               (str(handler[0]).split(' '))[2], self.name)

    def _finished(self):
        """Internal function to trigger finished state and related processing."""
        self._status = SoundInstanceStatus.finished

        # Call any registered finished handlers
        for handler in self._registered_finished_handlers:
            merged_kwargs = dict(list(handler[2].items()))
            merged_kwargs['sound_instance'] = self

            handler[0](**merged_kwargs)

    @property
    def status(self):
        """Return the current status of the sound instance."""
        return self._status

    @property
    def played(self):
        """Return whether or not this sound instance has been played."""
        return self._played

    @property
    def finished(self):
        """Return whether or not this sound instance has finished playing."""
        return self._status == SoundInstanceStatus.finished

    @property
    def loop_count(self):
        """Return how many times this sound instance has looped back to the beginning."""
        return self._loop_count


class DuckingSettings:
    """ DuckingSettings contains the parameters needed to control audio ducking
    for a sound.
    """

    def __init__(self, mc, config):
        """
        Constructor
        Args:
            mc: The media controller instance.
            config: The ducking configuration file section that contains all the ducking
                settings for the sound.

        Notes:
            The config section should contain the following attributes:
                target: A list of track names to apply the ducking to when the sound is played.
                delay: The duration (in seconds) of the delay period (time before attack starts)
                attack: The duration (in seconds) of the attack stage of the ducking envelope
                attenuation: The attenuation (gain) (0.0 to 1.0) to apply to the target track while
                    ducking
                release_point: The point (in seconds) relative to the end of the sound at which
                    to start the release stage.  A positive value indicates prior to the end of
                    the sound while a negative value indicates to start the release after the
                    end of the sound.
                release: The duration (in seconds) of the release stage of the ducking process.
        """
        if config is None:
            raise AudioException("The 'ducking' configuration must include the following "
                                 "attributes: track, delay, attack, attenuation, "
                                 "release_point, and release")

        if 'target' not in config:
            raise AudioException("'ducking.target' must contain at least one "
                                 "valid audio track name")

        # Target can contain a list of track names - convert string to list and validate
        self._targets = Util.string_to_event_list(config['target'])
        if not self._targets:
            raise AudioException("'ducking.target' must contain at least one "
                                 "valid audio track name")

        # Create a bit mask of target tracks based on their track number (will be used to pass
        # the target data to the audio library).
        self._track_bit_mask = 0
        for target in self._targets:
            track = mc.sound_system.audio_interface.get_track_by_name(target)
            if track is None:
                raise AudioException("'ducking.target' contains an invalid track name '{}'".format(target))
            self._track_bit_mask += (1 << track.number)

        # Delay is optional (defaults to 0, must be >= 0)
        if 'delay' in config:
            self._delay = max(mc.sound_system.audio_interface.string_to_secs(
                config['delay']), 0)
        else:
            self._delay = 0

        if 'attack' not in config:
            raise AudioException("'ducking.attack' must contain a valid attack value (time "
                                 "string)")
        self._attack = max(mc.sound_system.audio_interface.string_to_secs(config['attack']),
                           mc.sound_system.audio_interface.string_to_secs(
                               MINIMUM_DUCKING_DURATION))

        if 'attenuation' not in config:
            raise AudioException("'ducking.attenuation' must contain valid attenuation "
                                 "value (0.0 to 1.0)")
        self._attenuation = min(max(float(AudioInterface.string_to_gain(
            config['attenuation'])), 0.0), 1.0)

        if 'release_point' not in config:
            raise AudioException("'ducking.release_point' must contain a valid release point "
                                 "value (time string)")
        # Release point cannot be negative (must be before or at the end of the sound)
        self._release_point = max(mc.sound_system.audio_interface.string_to_secs(
            config['release_point']), 0)

        if 'release' not in config:
            raise AudioException("'ducking.release' must contain a valid release "
                                 "value (time string)")
        self._release = max(mc.sound_system.audio_interface.string_to_secs(config['release']),
                            mc.sound_system.audio_interface.string_to_secs(
                                MINIMUM_DUCKING_DURATION))

    @property
    def targets(self):
        """Return the list of target track names"""
        return self._targets

    @property
    def track_bit_mask(self):
        """ A bit mask of target tracks (each bit represents a track)"""
        return self._track_bit_mask

    @property
    def delay(self):
        """The duration (in samples) of the delay period (time before attack starts)"""
        return self._delay

    @property
    def attack(self):
        """The duration (in samples) of the attack stage of the ducking envelope"""
        return self._attack

    @property
    def attenuation(self):
        """The attenuation (gain) (0.0 to 1.0) to apply to the target track while ducking"""
        return self._attenuation

    @property
    def release_point(self):
        """The point (in samples) relative to the end of the sound at which
        to start the release stage.  A positive value indicates prior to the end of
        the sound while a negative value indicates to start the release after the
        end of the sound.
        """
        return self._release_point

    @property
    def release(self):
        """The duration (in samples) of the release stage of the ducking process."""
        return self._release

    def validate(self, sound_length):
        """
        Validates the ducking settings against the length of the sound to ensure all
        settings are valid.
        Args:
            sound_length: The length of the sound in samples

        Return:
            True if all settings are valid, otherwise an exception will be thrown
        """
        if sound_length is None or sound_length == 0:
            raise AudioException("ducking may not be applied to an empty/zero length sound")

        if self._attack > sound_length:
            raise AudioException("'ducking.attack' value may not be longer than the "
                                 "length of the sound")

        if self._release_point >= sound_length:
            raise AudioException("'ducking.release_point' value may not occur before the "
                                 "beginning of the sound")

        if self._release_point + self._attack >= sound_length:
            raise AudioException("'ducking.release_point' value may not occur before "
                                 "the ducking attack segment has completed")

        return True



from kivy.uix.effectwidget import EffectBase
from kivy.properties import ListProperty


class MonochromeEffect(EffectBase):
    """GLSL effect to convert the texture to monochrome.

    More information here:
    http://www.johndcook.com/blog/2009/08/24/algorithms-convert-color-grayscale/

    """

    luminosity = ListProperty([.299, .587, .114])
    '''This defines the luminosity factor for each color channel. The value
    for each channel must be between 0.0 and 1.0.

    :attr:`luminosity` is a :class:`ListProperty` defaults to
    (.299, .587, .114)
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_luminosity(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = monochrome_glsl.format(float(self.luminosity[0]),
                                           float(self.luminosity[1]),
                                           float(self.luminosity[2]))


monochrome_glsl = '''
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    float lum = ((color.x * {}) + (color.y * {}) + (color.z * {}));
    return vec4(lum, lum, lum, 1.0);
}}
'''

effect_cls = MonochromeEffect
name = 'monochrome'



from mpf.config_players.bcp_plugin_player import BcpPluginPlayer


class DisplayLightPlayer(BcpPluginPlayer):

    config_file_section = 'display_light_player'
    show_section = 'display_lights'

    def __init__(self, machine):
        super().__init__(machine)

        self.machine.events.add_handler("display_light_player_apply", self._apply_lights)

    def _apply_lights(self, context, element, values, **kwargs):
        del kwargs
        context_dict = self._get_instance_dict(context)
        if element not in context_dict:
            return

        # get delta to last frame for interframe interpolation
        last_frame = context_dict[element][0]
        current_frame = self.machine.clock.get_time()
        context_dict[element][0] = current_frame

        priority = context_dict[element][1]

        key = "display_light_player_{}".format(element)
        fade_ms = int((current_frame - last_frame) * 1000)
        for light, color in values.items():
            if color == -1:
                self.machine.lights[light].remove_from_stack_by_key(key=key, fade_ms=fade_ms)
            else:
                self.machine.lights[light].color(key=key, color=color, priority=priority, fade_ms=fade_ms)

    def _validate_config_item(self, device, device_settings):
        device_settings = super()._validate_config_item(device, device_settings)

        for device, settings in device_settings.items():
            settings['light_map'] = self._build_light_map(settings['lights'])

        return device_settings

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Trigger remote player via BCP."""
        context_dict = self._get_instance_dict(context)

        for element, s in settings.items():
            if s['action'] == "play":
                context_dict[element] = [self.machine.clock.get_time(), priority]
            elif s['action'] == "stop":
                try:
                    del context_dict[element]
                except IndexError:
                    pass
                self._clear_key_from_lights(element)
            else:
                raise AssertionError("Invalid action {}".format(s['action']))

        super().play(settings, context, calling_context, priority, **kwargs)

    def _clear_key_from_lights(self, element):
        # remove color from leds
        key = "display_light_player_{}".format(element)
        for light in self.machine.lights.values():
            light.remove_from_stack_by_key(key=key)

    def _build_light_map(self, tags):
        light_map = []
        for tag in tags:
            lights = self.machine.lights.items_tagged(tag) if tag != '*' else self.machine.lights
            for light in lights:
                if not light.config['x'] or not light.config['y']:
                    continue
                if light.config['x'] < 0 or light.config['x'] > 1.0:
                    raise AssertionError("x needs to be between 0 and 1 for light {}".format(light.name))
                if light.config['y'] < 0 or light.config['y'] > 1.0:
                    raise AssertionError("y needs to be between 0 and 1 for light {}".format(light.name))

                light_map.append((light.config['x'], light.config['y'], light.name))

        return light_map

    def clear_context(self, context):
        """Clean up all active lights."""
        context_dics = self._get_instance_dict(context)
        for element in context_dics:
            self._clear_key_from_lights(element)

        self._reset_instance_dict(context)

        super().clear_context(context)


def register_with_mpf(machine):
    """Register widget player in MPF module."""
    return 'display_light', DisplayLightPlayer(machine)


player_cls = DisplayLightPlayer



import importlib
import abc
from typing import Optional, List, Union

from kivy.event import EventDispatcher
from kivy.uix.effectwidget import (MonochromeEffect, InvertEffect,
                                   ScanlinesEffect, ChannelMixEffect,
                                   PixelateEffect, HorizontalBlurEffect,
                                   VerticalBlurEffect, FXAAEffect,
                                   EffectBase)

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class EffectsManager:
    def __init__(self, mc: "MpfMc") -> None:
        self.mc = mc
        self._effects = dict()
        self._register_mpf_effects()
        self._register_kivy_effects()

    @property
    def effects(self) -> dict:
        """Return effects."""
        return self._effects

    def register_effect(self, name: str, transition_cls) -> None:
        """Register effect."""
        self._effects[name] = transition_cls

    def get_effect(self, config: Optional[dict] = None) -> List["EffectBase"]:
        """Return effects."""
        if config:
            # The kivy shader transitions can't accept unexpected kwargs
            kwargs = config.copy()
            effect_obj = self._effects[config['type']]()

            # Set effect properties
            for attr, value in kwargs.items():
                if hasattr(effect_obj, attr):
                    setattr(effect_obj, attr, value)

            if isinstance(effect_obj, EffectBase):
                return [effect_obj]
            elif isinstance(effect_obj, EffectsChain):
                return effect_obj.get_effects()
            else:
                return []
        else:
            return []

    def _register_mpf_effects(self) -> None:
        for t in self.mc.machine_config['mpf-mc']['mpf_effect_modules']:
            i = importlib.import_module('mpfmc.effects.{}'.format(t))
            self.register_effect(getattr(i, 'name'),
                                 getattr(i, 'effect_cls'))

    def _register_kivy_effects(self) -> None:
        self.register_effect('invert_colors', InvertEffect)
        self.register_effect('scanlines', ScanlinesEffect)
        self.register_effect('color_channel_mix', ChannelMixEffect)
        self.register_effect('pixelate', PixelateEffect)
        self.register_effect('horizontal_blur', HorizontalBlurEffect)
        self.register_effect('vertical_blur', VerticalBlurEffect)
        self.register_effect('anti_aliasing', FXAAEffect)

    def validate_effects(self, config: Union[dict, list]) -> list:
        """Validate the effects section of a widget.

        Args:
            config: The localized 'effects' config dictionary for a single widget.

        Returns:
            A list of 'effects' config dictionary entries that have been validated.
        """
        if isinstance(config, dict):
            config = [config]

        effect_list = list()

        for effect in config:
            effect_list.append(self.process_effect(effect))

        return effect_list

    def process_effect(self, config: dict) -> dict:
        try:
            effect_cls = self._effects[config['type']]
            del effect_cls
        except (KeyError, TypeError):
            try:
                raise ValueError(
                    '"{}" is not a valid MPF widget effect type. Did you '
                    'misspell it, or forget to enable it in the "mpf-mc: '
                    'mpf_effect_modules" section of your machine config?'.format(
                        config['type']))
            except (KeyError, TypeError):
                raise ValueError("Invalid widget effects config: {}".format(config))

        self.mc.config_validator.validate_config('effects:{}'.format(
            config['type']).lower(), config, base_spec='effects:common')

        return config


class EffectsChain(EventDispatcher, metaclass=abc.ABCMeta):

    """Abstract base class for an effect that is actually a chain of
    one or more individual effects."""

    @abc.abstractmethod
    def get_effects(self) -> List["EffectBase"]:
        """Return the list of effects in this chain."""
        raise NotImplementedError('get_effects method must be defined to use this base class')



"""A triangle widget."""
from typing import Optional

from kivy.graphics import Triangle as KivyTriangle
from kivy.graphics.context_instructions import Color, Rotate, Scale
from kivy.properties import ListProperty, NumericProperty

from mpfmc.uix.widget import Widget
from mpfmc.core.utils import center_of_points_list

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Triangle(Widget):

    """A triangle widget."""

    widget_type_name = 'Triangle'
    animation_properties = ('points', 'color', 'opacity', 'rotation', 'scale')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        super().__init__(mc=mc, config=config, key=key)

        # The points in this widget are always relative to the bottom left corner
        self.anchor_pos = ("left", "bottom")

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(color=self._draw_widget,
                  points=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget)

        self._draw_widget()

    def _draw_widget(self, *args) -> None:
        """Establish the drawing instructions for the widget."""
        del args

        if self.canvas is None:
            return

        # TODO: allow user to set rotation/scale origin
        center = center_of_points_list(self.points)
        self.canvas.clear()

        with self.canvas:
            Color(*self.color)
            Scale(self.scale, origin=center)
            Rotate(angle=self.rotation, origin=center)
            KivyTriangle(points=self.points)

    #
    # Properties
    #

    points = ListProperty([0, 0, 50, 100, 100, 0])
    '''The list of points to use to draw the widget in (x1, y1, x2, y2,
    x3, y3) format.

    :attr:`points` is a :class:`~kivy.properties.ListProperty`.
    '''

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [Triangle]



from kivy.properties import ListProperty
from kivy.uix.effectwidget import EffectBase


class ColorizeEffect(EffectBase):
    """GLSL effect to apply a color tint to a texture."""

    tint_color = ListProperty([1, 0.4, 0, 0])
    '''This defines the color of the tint to be used in the effect.

    :attr:`tint_color` is a :class:`ListProperty` defaults to
    (1, 0.4, 1, 0)
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_tint_color(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = colorize_glsl.format(float(self.tint_color[0]),
                                         float(self.tint_color[1]),
                                         float(self.tint_color[2]))


colorize_glsl = '''
        vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
        {{
            vec4 c = vec4(color.x * {}, color.y * {}, color.z * {}, 1.0);
            return c;
        }}
        '''

effect_cls = ColorizeEffect
name = 'colorize'



from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.sound_file cimport *
from mpfmc.core.audio.track cimport *
from mpfmc.core.audio.notification_message cimport *


# ---------------------------------------------------------------------------
#    Sound Loop Track types
# ---------------------------------------------------------------------------
cdef enum:
    do_not_stop_loop = 0xFFFFFFFF

cdef enum LayerStatus:
    layer_stopped = 0
    layer_queued = 1
    layer_playing = 2
    layer_fading_in = 3
    layer_fading_out = 4

ctypedef struct SoundLoopLayerSettings:
    LayerStatus status
    SoundSample *sound
    Uint8 volume
    long sound_loop_set_id
    Uint64 sound_id
    Uint32 fade_in_steps
    Uint32 fade_out_steps
    Uint32 fade_steps_remaining
    bint looping
    Uint8 marker_count
    GArray *markers

cdef enum SoundLoopSetPlayerStatus:
    # Enumeration of the possible sound loop set player status values.
    player_idle = 0
    player_delayed = 1
    player_fading_in = 2
    player_fading_out = 3
    player_playing = 4

ctypedef struct SoundLoopSetPlayer:
    SoundLoopSetPlayerStatus status
    Uint32 length
    SoundLoopLayerSettings master_sound_layer
    GSList *layers     # An array of SoundLoopLayerSettings objects
    Uint32 sample_pos
    Uint32 stop_loop_samples_remaining
    Uint32 start_delay_samples_remaining
    float tempo

ctypedef struct TrackSoundLoopState:
    # State variables for TrackSoundLoop tracks
    GSList *players
    SoundLoopSetPlayer *current


# ---------------------------------------------------------------------------
#    TrackSoundLoop class
# ---------------------------------------------------------------------------
cdef class TrackSoundLoop(Track):

    # Track state needs to be stored in a C struct in order for them to be accessible in
    # the SDL callback functions without the GIL (for performance reasons).
    # The TrackSoundLoopState struct is allocated during construction and freed during
    # destruction.
    cdef TrackSoundLoopState *type_state
    cdef long _sound_loop_set_counter
    cdef dict _active_sound_loop_sets

    cdef process_notification_message(self, NotificationMessageContainer *notification_message)

    cdef _apply_layer_settings(self, SoundLoopLayerSettings *layer, dict layer_settings)
    cdef _initialize_player(self, SoundLoopSetPlayer *player)
    cdef _delete_player(self, SoundLoopSetPlayer *player)
    cdef _delete_player_layers(self, SoundLoopSetPlayer *player)
    cdef _cancel_all_delayed_players(self)
    cdef _fade_out_all_players(self, Uint32 fade_steps)
    cdef inline Uint32 _round_sample_pos_up_to_interval(self, Uint32 sample_pos, Uint32 interval, int bytes_per_sample_frame)

    @staticmethod
    cdef void mix_playing_sounds(TrackState *track, Uint32 buffer_length, AudioCallbackData *callback_data) nogil


cdef SoundLoopLayerSettings *_create_sound_loop_layer_settings() nogil



"""Threaded Asset Loader for MC."""
import logging
import threading
import traceback
from queue import PriorityQueue, Queue, Empty

import sys

from mpf.core.assets import BaseAssetManager


class ThreadedAssetManager(BaseAssetManager):

    """AssetManager which uses the Threading module."""

    def __init__(self, machine):
        """Initialise queues and start loader thread."""
        super().__init__(machine)
        self.loader_queue = PriorityQueue()  # assets for to the loader thread
        self.loaded_queue = Queue()  # assets loaded from the loader thread
        self.loader_thread = None
        self._loaded_watcher = False

        self._start_loader_thread()

    def _start_loader_thread(self):
        self.loader_thread = AssetLoader(loader_queue=self.loader_queue,
                                         loaded_queue=self.loaded_queue,
                                         exception_queue=self.machine.crash_queue,
                                         thread_stopper=self.machine.thread_stopper)
        self.loader_thread.daemon = True
        self.loader_thread.start()

    def load_asset(self, asset):
        """Put asset in loader queue."""
        # Internal method which handles the logistics of actually loading an
        # asset. Should only be called by Asset.load() as that method does
        # additional things that are needed.

        self.num_assets_to_load += 1

        # It's ok for an asset to make it onto this queue twice as the loader
        # thread will check the asset's loaded attribute to make sure it needs
        # to load it.

        # This is a PriorityQueue which will automatically put the asset into
        # the proper position in the queue based on its priority.

        self.loader_queue.put(asset)

        if not self._loaded_watcher:
            self._loaded_watcher = self.machine.clock.schedule_interval(self._check_loader_status, 0)

    def _check_loader_status(self, *args):
        del args
        # checks the loaded queue and updates loading stats
        try:
            while not self.loaded_queue.empty():
                asset, loaded = self.loaded_queue.get()
                if loaded:
                    asset.is_loaded()
                self.num_assets_loaded += 1
                self._post_loading_event()
        except AttributeError:
            pass

        if self.num_assets_to_load == self.num_assets_loaded:
            self.num_assets_loaded = 0
            self.num_assets_to_load = 0
            self.machine.clock.unschedule(self._loaded_watcher)
            self._loaded_watcher = None


class AssetLoader(threading.Thread):

    """Base class for the Asset Loader thread and actually loads the assets from disk.

    Args:
        loader_queue: A reference to the asset manager's loader_queue which
            holds assets waiting to be loaded. Items are automatically sorted
            in reverse order by priority, then creation ID.
        loaded_queue: A reference to the asset manager's loaded_queue which
            holds assets that have just been loaded. Entries are Asset
            instances.
        exception_queue: Send a reference to self.machine.crash_queue. This way if
            the asset loader crashes, it will write the crash to that queue and
            cause an exception in the main thread. Otherwise it fails silently
            which is super annoying. :)
    """

    def __init__(self, loader_queue, loaded_queue, exception_queue,
                 thread_stopper):
        """Initialise asset loader."""
        threading.Thread.__init__(self)
        self.log = logging.getLogger('Asset Loader')
        self.loader_queue = loader_queue
        self.loaded_queue = loaded_queue
        self.exception_queue = exception_queue
        self.thread_stopper = thread_stopper
        self.name = 'asset_loader'

    def run(self):
        """Run loop for the loader thread."""
        try:  # wrap the so we can send exceptions to the main thread
            while not self.thread_stopper.is_set():
                try:
                    asset = self.loader_queue.get(block=True, timeout=1)
                except Empty:
                    asset = None

                if asset:
                    with asset.lock:
                        if not asset.loaded:
                            asset.do_load()
                            self.loaded_queue.put((asset, True))
                        else:
                            self.loaded_queue.put((asset, False))

            return

        # pylint: disable-msg=broad-except
        except Exception:  # pragma: no cover
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value,
                                               exc_traceback)
            msg = ''.join(line for line in lines)
            self.exception_queue.put(msg)
            raise



import os
import sys
import unittest

from mpf.tests.MpfTestCase import UnitTestConfigLoader

verbose = sys.argv and "-v" in sys.argv

if not verbose:
    os.environ['KIVY_NO_FILELOG'] = '1'
    os.environ['KIVY_NO_CONSOLELOG'] = '1'

from kivy.graphics.opengl import glReadPixels, GL_RGB, GL_UNSIGNED_BYTE
from kivy import Config, Logger
from kivy.base import runTouchApp, stopTouchApp, EventLoop
from kivy.clock import Clock
from kivy.uix.widget import Widget as KivyWidget

import mpfmc
from mpf.core.utility_functions import Util

if not verbose:
    Config.set('kivy', 'log_enable', '0')
    Config.set('kivy', 'log_level', 'warning')

from mpfmc.core.mc import MpfMc
from time import time, sleep

sys.stderr = sys.__stderr__


class MpfMcTestCase(unittest.TestCase):
    def __init__(self, *args):
        self.sent_bcp_commands = list()
        super().__init__(*args)

        self._events = dict()
        self._last_event_kwargs = dict()
        self.max_test_setup_secs = 30

        self._fps = 30

    def _mc_time(self):
        return self._current_time

    def get_options(self):
        return dict(machine_path=self.get_machine_path(),
                    mcconfigfile='mcconfig.yaml',
                    production=False,
                    configfile=Util.string_to_event_list(self.get_config_file()),
                    no_load_cache=False,
                    create_config_cache=True,
                    no_sound=False,
                    bcp=False)

    def get_absolute_machine_path(self):
        return os.path.abspath(os.path.join(
            mpfmc.__path__[0], os.pardir, 'mpfmc', self.get_machine_path()))

    def get_machine_path(self):
        raise NotImplementedError

    def get_config_file(self):
        raise NotImplementedError

    def get_abs_path(self, path):
        return os.path.join(os.path.abspath(os.curdir), path)

    def advance_time(self, secs=.1):
        start = self._current_time
        while self._current_time < start + secs:
            EventLoop.idle()
            self._current_time += 1 / self._fps

    def advance_real_time(self, secs=.1):
        start = self._current_time
        while self._current_time < start + secs:
            EventLoop.idle()
            sleep(1 / self._fps)
            self._current_time += 1 / self._fps

        EventLoop.idle()

    def get_pixel_color(self, x, y):
        """Returns a binary string of the RGB bytes that make up the slide
        pixel at the passed x,y coordinates. 2 bytes per pixel, 6 bytes total.
        This method *does* compensate for different window sizes.

        Note: This method does not work yet.

        """
        raise NotImplementedError  # remove when we fix it. :)

        # do the Window import here because we don't want to import it at the
        # top or else we won't be able to set window properties
        from kivy.core.window import Window

        # convert the passed x/y to the actual x/y of the Window since it's
        # possible for the mpf-mc display size to be different than the Window
        # size
        x *= Window.width / Window.children[0].width
        y *= Window.height / Window.children[0].height

        return glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE)

    def tearDown(self):
        self.mc.stop()

    def patch_bcp(self):
        # used internally
        self.orig_bcp_send = self.mc.bcp_processor.send
        self.mc.bcp_processor.send = self._bcp_send

        # this is used to send BCP commands to mpf-mc
        self.send = self.mc.bcp_processor._process_command

        self.mc.bcp_client_connected = True

    def _bcp_send(self, bcp_command, callback=None, **kwargs):
        # used for commands sent from the MC to the PC
        # print((bcp_command, callback, kwargs))
        self.sent_bcp_commands.append((bcp_command, callback, kwargs))
        self.orig_bcp_send(bcp_command=bcp_command, callback=callback,
                           **kwargs)

    def setUp(self):
        # Most of the setup is done in run(). Explanation is there.
        Config._named_configs.pop('app', None)

        self._start_time = time()
        self._current_time = self._start_time
        Clock._start_tick = self._start_time
        Clock._last_tick = self._start_time
        Clock.time = self._mc_time

        # prevent sleep in clock
        Clock._max_fps = 0
        # reset clock
        Clock._root_event = None

        from mpf.core.player import Player
        Player.monitor_enabled = False

        machine_path = self.get_absolute_machine_path()

        # load config
        config_loader = UnitTestConfigLoader(machine_path, [self.get_config_file()], {}, {}, {})

        config = config_loader.load_mc_config()

        try:
            self.mc = MpfMc(options=self.get_options(), config=config)

            self.patch_bcp()

            from kivy.core.window import Window
            Window.create_window()
            Window.canvas.clear()

            self._start_app_as_slave()
        except Exception:
            if hasattr(self, "mc") and self.mc:
                # prevent dead locks with two asset manager threads
                self.mc.stop()
            raise

    def _start_app_as_slave(self):
        # from app::run
        if not self.mc.built:
            self.mc.load_config()
            self.mc.load_kv(filename=self.mc.kv_file)
            root = self.mc.build()
            if root:
                self.mc.root = root
        if self.mc.root:
            if not isinstance(self.mc.root, KivyWidget):
                Logger.critical('App.root must be an _instance_ of Kivy Widget')
                raise Exception('Invalid instance in App.root')
            from kivy.core.window import Window
            Window.add_widget(self.mc.root)

        # Check if the window is already created
        from kivy.base import EventLoop
        window = EventLoop.window
        if window:
            self.mc._app_window = window
            #window.set_title(self.mc.get_application_name() + self._testMethodName)
            icon = self.mc.get_application_icon()
            if icon:
                window.set_icon(icon)
            self.mc._install_settings_keys(window)
        else:
            Logger.critical("Application: No window is created."
                            " Terminating application run.")
            return

        self.mc.dispatch('on_start')
        runTouchApp(slave=True)  # change is here

        # Perform init process
        tries = 0
        while not self.mc.is_init_done.is_set() and not self.mc.thread_stopper.is_set():
            self.advance_time()
            sleep(.001)
            tries += 1
            if tries > 1000:
                self.fail("Test init took too long")

        # set a nice title
        window.set_title(self.__class__.__name__ + "::" + self._testMethodName)

    def dump_clock(self):
        print("---------")
        events = []
        event = Clock._root_event
        while event:
            events.append(event)
            event = event.next

        events.sort(key=lambda x: str(x.get_callback()))

        for event in events:
            print(event.get_callback(), event.timeout)

    def _mock_event_handler(self, event_name, **kwargs):
        self._last_event_kwargs[event_name] = kwargs
        self._events[event_name] += 1

    def mock_event(self, event_name):
        self._events[event_name] = 0
        self.mc.events.remove_handler_by_event(
            event=event_name, handler=self._mock_event_handler)
        self.mc.events.add_handler(event=event_name,
                                   handler=self._mock_event_handler,
                                   event_name=event_name)

    def assertEventNotCalled(self, event_name):
        """Assert that event was not called."""
        if event_name not in self._events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._events[event_name] != 0:
            raise AssertionError("Event {} was called {} times.".format(
                event_name, self._events[event_name]))

    def assertEventCalled(self, event_name, times=None):
        """Assert that event was called."""
        if event_name not in self._events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._events[event_name] == 0:
            raise AssertionError("Event {} was not called.".format(event_name))

        if times is not None and self._events[event_name] != times:
            raise AssertionError("Event {} was called {} instead of {}.".format(
                event_name, self._events[event_name], times))

    def assertEventCalledWith(self, event_name, **kwargs):
        """Assert that event was called with kwargs."""
        self.assertEventCalled(event_name)
        self.assertEqual(kwargs, self._last_event_kwargs[event_name],
                         "Args for {} differ.".format(event_name))

    def reset_mock_events(self):
        for event in self._events.keys():
            self._events[event] = 0



from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.track cimport TrackState


# ---------------------------------------------------------------------------
#    Notification Message types
# ---------------------------------------------------------------------------

cdef enum NotificationMessage:
    notification_sound_started = 1            # Notification that a sound has started playing
    notification_sound_stopped = 2            # Notification that a sound has stopped
    notification_sound_looping = 3            # Notification that a sound is looping back to the beginning
    notification_sound_marker = 4             # Notification that a sound marker has been reached during playback
    notification_sound_about_to_finish = 5    # Notification that a sound is about to finish playing
    notification_player_idle = 10             # Notification that a player is now idle and ready to play another sound
    notification_track_stopped = 0            # Notification that the track has stopped
    notification_track_paused = 21            # Notification that the track has been paused
    notification_sound_loop_set_started = 31  # Notification that a sound_loop_set has started playing
    notification_sound_loop_set_stopped = 32  # Notification that a sound_loop_set has stopped
    notification_sound_loop_set_looping = 33  # Notification that a sound_loop_set is looping back to the beginning

ctypedef struct NotificationMessageDataLooping:
    int loop_count
    int loops_remaining

ctypedef struct NotificationMessageDataMarker:
    int id

ctypedef struct NotificationMessageSoundLoopSet:
    long id
    gpointer player


ctypedef union NotificationMessageData:
    NotificationMessageDataLooping looping
    NotificationMessageDataMarker marker
    NotificationMessageSoundLoopSet sound_loop_set

ctypedef struct NotificationMessageContainer:
    NotificationMessage message
    Uint64 sound_id
    Uint64 sound_instance_id
    int player
    NotificationMessageData data


# ---------------------------------------------------------------------------
#    Notification Message functions
# ---------------------------------------------------------------------------

cdef inline NotificationMessageContainer *_create_notification_message() nogil:
    """
    Creates a new notification message.
    :return: A pointer to the new notification message.
    """
    return <NotificationMessageContainer*>g_slice_alloc0(sizeof(NotificationMessageContainer))

cdef inline void send_sound_started_notification(int player, Uint64 sound_id, Uint64 sound_instance_id,
                                                 TrackState *track) nogil:
    """
    Sends a sound started notification
    Args:
        player: The sound player number on which the event occurred
        sound_id: The sound id
        sound_instance_id: The sound instance id
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_started
        notification_message.player = player
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = sound_instance_id

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_stopped_notification(int player, Uint64 sound_id, Uint64 sound_instance_id,
                                                 TrackState *track) nogil:
    """
    Sends a sound stopped notification
    Args:
        player: The sound player number on which the event occurred
        sound_id: The sound id
        sound_instance_id: The sound instance id
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_stopped
        notification_message.player = player
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = sound_instance_id

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_looping_notification(int player, Uint64 sound_id, Uint64 sound_instance_id,
                                                 TrackState *track) nogil:
    """
    Sends a sound looping notification
    Args:
        player: The sound player number on which the event occurred
        sound_id: The sound id
        sound_instance_id: The sound instance id
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_looping
        notification_message.player = player
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = sound_instance_id

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_about_to_finish_notification(int player, Uint64 sound_id, Uint64 sound_instance_id,
                                                         TrackState *track) nogil:
    """
    Sends a sound about to finish notification
    Args:
        player: The sound player number on which the event occurred
        sound_id: The sound id
        sound_instance_id: The sound instance id
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_about_to_finish
        notification_message.player = player
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = sound_instance_id

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_marker_notification(int player, Uint64 sound_id, Uint64 sound_instance_id,
                                                TrackState *track,
                                                int marker_id) nogil:
    """
    Sends a sound marker notification message
    Args:
        player: The sound player number on which the event occurred
        sound_id: The sound id
        sound_instance_id: The sound instance id
        track: The TrackState pointer
        marker_id: The id of the marker being sent for the specified sound
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_marker
        notification_message.player = player
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = sound_instance_id
        notification_message.data.marker.id = marker_id

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_track_stopped_notification(TrackState *track) nogil:
    """
    Sends a track stopped notification
    Args:
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_track_stopped
        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_track_paused_notification(TrackState *track) nogil:
    """
    Sends a track paused notification
    Args:
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_track_paused
        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_loop_set_started_notification(int sound_loop_set_id, Uint64 sound_id, gpointer player, TrackState *track) nogil:
    """
    Sends a sound_loop_set started notification
    Args:
        sound_loop_set_id: The sound_loop_set id
        sound_id: The sound id
        player: A pointer to the sound loop player that was playing this sound loop set 
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_loop_set_started
        notification_message.player = 0
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = 0
        notification_message.data.sound_loop_set.id = sound_loop_set_id
        notification_message.data.sound_loop_set.player = player

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_loop_set_stopped_notification(int sound_loop_set_id, Uint64 sound_id, gpointer player, TrackState *track) nogil:
    """
    Sends a sound_loop_set stopped notification
    Args:
        sound_loop_set_id: The sound_loop_set id
        sound_id: The sound id
        player: A pointer to the sound loop player that was playing this sound loop set 
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_loop_set_stopped
        notification_message.player = 0
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = 0
        notification_message.data.sound_loop_set.id = sound_loop_set_id
        notification_message.data.sound_loop_set.player = player

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)

cdef inline void send_sound_loop_set_looping_notification(int sound_loop_set_id, Uint64 sound_id, gpointer player, TrackState *track) nogil:
    """
    Sends a sound_loop_set looping notification
    Args:
        sound_loop_set_id: The sound_loop_set id
        sound_id: The sound id
        player: A pointer to the sound loop player that was playing this sound loop set 
        track: The TrackState pointer
    """
    cdef NotificationMessageContainer *notification_message = _create_notification_message()
    if notification_message != NULL:
        notification_message.message = notification_sound_loop_set_looping
        notification_message.player = 0
        notification_message.sound_id = sound_id
        notification_message.sound_instance_id = 0
        notification_message.data.sound_loop_set.id = sound_loop_set_id
        notification_message.data.sound_loop_set.player = player

        track.notification_messages = g_slist_prepend(track.notification_messages, notification_message)




import logging
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAudioBadBufferSetting(MpfMcTestCase):
    """
    Tests the audio settings with a bad buffer setting (not a power of two)
    """

    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_bad_buffer_setting.yaml'

    def test_default_sound_system(self):
        """ Tests the sound system and audio interface with default settings """

        if self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - unable to run audio tests")
            self.skipTest("Sound system is not enabled - unable to run audio tests")

        logging.getLogger('TestAudio').setLevel(10)

        self.assertIsNotNone(self.mc.sound_system)

        if self.mc.sound_system.audio_interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(self.mc.sound_system.audio_interface)
        settings = self.mc.sound_system.audio_interface.get_settings()
        self.assertIsNotNone(settings)
        self.assertIn(settings['buffer_samples'], [1024, 2048])
        self.assertIn(settings['audio_channels'], [1, 2])
        self.assertEqual(settings['sample_rate'], 44100)



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestLine(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/shapes'

    def get_config_file(self):
        return 'test_shapes.yaml'

    def test_line(self):
        self.mc.events.post('slide1')
        self.advance_time()



#!python
#cython: embedsignature=True, language_level=3

import logging
from functools import partial

from mpfmc.config_collections.playlist import PlaylistInstance
from mpfmc.core.audio.track cimport *


# ---------------------------------------------------------------------------
#    PlaylistController class
# ---------------------------------------------------------------------------
class PlaylistController:
    """
    PlaylistController class
    """

    def __init__(self, object mc, object track, float crossfade_time=0.0, float volume=1.0):
        """
        Constructor
        Args:
            mc: The media controller app
            track: The audio track that will be managed by this playlist controller (standard track)
            crossfade_time: The default crossfade time (secs) to use when fading between sounds
            volume: The playlist volume (0.0 to 1.0)
        """
        self.log = logging.getLogger("PlaylistController")
        self.mc = mc
        self._track = track
        self._crossfade_time = crossfade_time

        # Dictionary of PlaylistInstance class objects keyed by SoundInstance ID
        self._playlists_by_sound_instance = dict()

        # Keep track of current and previous playlist
        self._current_playlist = None

        # Keeps track of a pending request that cannot be immediately serviced because the
        # underlying audio track is busy (only 2 players)
        self._pending_request = None

        # Create event handlers for sound instance events
        # Note: we don't have to worry about played events since the playlist controller will initiate
        # playback and therefore will already know when those events are occurring.
        self.mc.events.add_handler('{}_playlist_sound_stopped'.format(self.name.lower()),
                                   self._on_sound_instance_stopped, 0)
        self.mc.events.add_handler('{}_playlist_sound_about_to_finish'.format(self.name.lower()),
                                   self._on_sound_instance_about_to_finish)

        self.log.debug("Created PlaylistController %s: ", self._track.name)

    def __dealloc__(self):
        """Destructor"""
        pass

    def __repr__(self):
        return '<PlaylistController.{}>'.format(self.name)

    @property
    def name(self):
        """The name of the playlist controller (and corresponding audio track)"""
        return self._track.name

    @property
    def track(self):
        """The corresponding audio track the playlist is managing"""
        return self._track

    @property
    def crossfade_time(self):
        """The time (secs) to use when fading between sounds"""
        return self._crossfade_time

    @property
    def has_pending_request(self):
        """Whether the playlist controller has a pending request queued until it is not busy"""
        return self._pending_request is not None

    def stop_all_sounds(self, float fade_out_seconds = 0.0):
        """
        Stops all playing sounds immediately on the playlist track.
        Args:
            fade_out_seconds: The number of seconds to fade out the sounds before stopping
        """
        self._track.stop_all_sounds(fade_out_seconds)

        # TODO: playlist housekeeping (stopped events, remove instance, etc.)

    def play(self, str playlist not None, str context=None, dict player_settings=None):
        """
        Immediately play a playlist.

        Args:
            playlist: The name of the playlist asset object to play.
            context: The calling context (if any)
            player_settings: Settings to use for playback
        """

        self.log.debug("play - Preparing playlist '%s' for playback.", playlist)

        if player_settings is None:
            player_settings = dict()

        if playlist not in self.mc.playlists:
            self.log.error("PlaylistController (%s track): Could not play specified playlist "
                           "(%s) as it does not exist", self.name, playlist)
            return None

        # Is there already a previous playlist that is still active (fading)?
        if self._is_busy():
            # Delay play playlist until track is finished with current crossfade (too busy)
            self._pending_request = partial(self.play, playlist=playlist, context=context,
                                            player_settings=player_settings)

            self.log.debug("play - Playlist track is too busy. Delaying play playlist call.")
            return None

        # Determine settings (override playlist with player settings)
        playlist_instance = PlaylistInstance(playlist,
                                             self.mc.playlists[playlist],
                                             self.crossfade_time,
                                             context,
                                             player_settings)

        # Is there already a playlist playing?
        if self._current_playlist:
            # Already a playlist playing.  We know there is at least one free sound player because
            # of the _is_busy() call above.
            # Stop the current playlist
            self.stop()

        # Start the new playlist (now becomes the current playlist)
        self._current_playlist = playlist_instance

        # Post events when played for playlist
        for event in self._current_playlist.events_when_played:
            self.mc.post_mc_native_event(event)

        # Get the next sound to play from playlist
        sound_name = self._current_playlist.get_next_sound_name()
        self._play_playlist_sound(sound_name, self._current_playlist)

        return playlist_instance

    def _is_busy(self):
        """Returns whether or not all the sound players for the playlist track are currently busy"""
        if self._track.get_sound_players_in_use_count() > 1:
            return True
        else:
            return False

    def _play_playlist_sound(self, str sound_name, object playlist, float fade_in=0.0):
        """
        Plays the specified playlist sound
        Args:
            sound_name: The name of the sound to start playing
            playlist: The playlist from which the sound came from
            fade_in: The number of seconds over which to fade in the sound (frequently the
                crossfade time).

        """

        # Create sound instance
        sound = self.mc.sounds[sound_name]

        # Add custom events to post when particular sound actions occur (stop, about to finish).
        # These events are used to trigger playlist events (advance, stop)
        events_when_stopped = []
        if sound.events_when_stopped:
            events_when_stopped.extend(sound.events_when_stopped)
        events_when_stopped.extend(['{}_playlist_sound_stopped'.format(self.name.lower())])

        events_when_about_to_finish = []
        if sound.events_when_about_to_finish:
            events_when_about_to_finish.extend(sound.events_when_about_to_finish)
        events_when_about_to_finish.extend(['{}_playlist_sound_about_to_finish'.format(self.name.lower())])

        # Post events when sound changed as a new sound is playing now for the current playlist
        for event in self._current_playlist.events_when_sound_changed:
            self.mc.post_mc_native_event(event)

        # Play sound on playlist track (override certain settings needed to manage playlist)
        # Standard track will return a sound instance if play was successful
        sound_instance = self._track.play_sound(sound,
                                                playlist.context,
                                                {
                                                    'fade_in': fade_in,
                                                    'fade_out': 0.0,
                                                    'about_to_finish_time': playlist.crossfade_time,
                                                    'max_queue_time': None,
                                                    'events_when_stopped': events_when_stopped,
                                                    'events_when_about_to_finish': events_when_about_to_finish
                                                })

        # Associate sound instance with playlist instance
        self._playlists_by_sound_instance[sound_instance] = playlist
        playlist.current_sound_instance = sound_instance

    def stop(self):
        """Immediately stop the currently playing playlist. Will fade out using the crossfade setting."""
        if not self._current_playlist:
            self.log.debug("stop - No playlist is currently playing. Could not stop current playlist.")
            return

        self.log.debug("stop - Stopping the current playlist ('%s').",
                       self._current_playlist.name)

        # Stop the current sound (if another sound is fading out, let it finish on its own)
        if self._current_playlist.current_sound_instance:
            self._track.stop_sound_instance(self._current_playlist.current_sound_instance,
                                            self._current_playlist.crossfade_time)

        self._current_playlist = None

    def advance(self):
        """Advance the currently playing playlist to the next sound."""

        # If there is no current playlist, do nothing
        if not self._current_playlist:
            self.log.debug("advance - No playlist is currently playing. Could not advance to next sound")
            return

        self.log.debug("advance - Advancing the current playlist ('%s') to the next sound.",
                       self._current_playlist.name)

        if self._is_busy():
            # Delay advance playlist until track is finished with current crossfade (too busy)
            self._pending_request = partial(self.advance)
            self.log.debug("advance - Playlist track is too busy. Delaying advance to next sound")
            return

        # Determine if playlist will now repeat/loop.  Post playlist looping events (if necessary)
        if self._current_playlist.end_of_playlist:
            if self._current_playlist.repeat and self._current_playlist.events_when_looping:
                    for event in self._current_playlist.events_when_looping:
                        self.mc.post_mc_native_event(event)

        # Set the next sound in the sound player and calculate the fades based on the crossfade setting
        next_sound_name = self._current_playlist.get_next_sound_name()
        if next_sound_name:
            if self._current_playlist.current_sound_instance is not None:
                self._track.stop_sound_instance(self._current_playlist.current_sound_instance,
                                                self._current_playlist.crossfade_time)
                self._current_playlist.fading_sound_instance = self._current_playlist.current_sound_instance

            self._play_playlist_sound(next_sound_name, self._current_playlist, self._current_playlist.crossfade_time)

    def set_repeat(self, repeat=True):
        """Set whether or not the currently playing playlist should repeat when finished."""

        if self._current_playlist:
            self._current_playlist.loop = repeat
            self.log.debug("set_repeat - Setting repeat for currently playlist to {}.", str(repeat))
        else:
            self.log.debug("set_repeat - No playlist is currently playing. "
                           "Could not set repeat to {}.", str(repeat))

    def _on_sound_instance_stopped(self, sound_instance=None, **kwargs):
        """Callback function called whenever a playlist sound has finished playing."""

        if sound_instance is None or sound_instance not in self._playlists_by_sound_instance:
            return

        # Get playlist for sound_instance and remove it from dictionary of active sounds
        playlist = self._playlists_by_sound_instance[sound_instance]
        del self._playlists_by_sound_instance[sound_instance]

        if playlist.current_sound_instance == sound_instance:
            playlist.current_sound_instance = None
        if playlist.fading_sound_instance == sound_instance:
            playlist.fading_sound_instance = None

        # Post events when sound stopped for playlist
        for event in playlist.events_when_sound_stopped:
            self.mc.post_mc_native_event(event)

        # Playlist is finished when last sound instance of playlist has completed
        if playlist not in self._playlists_by_sound_instance.values():

            # Playlist has stopped

            # Trigger any stopped events
            if playlist.events_when_stopped is not None:
                for event in playlist.events_when_stopped:
                    self.mc.post_mc_native_event(event)

            if playlist == self._current_playlist:
                self._current_playlist = None

        # Service any pending request (play, advance)
        if self._pending_request:
            request = self._pending_request
            self._pending_request = None
            request()

    def _on_sound_instance_about_to_finish(self, sound_instance=None, **kwargs):
        """Callback function called whenever a playlist sound is about to finish playing."""

        if sound_instance is None or sound_instance not in self._playlists_by_sound_instance:
            return

        # Get playlist for sound_instance
        playlist = self._playlists_by_sound_instance[sound_instance]

        # Take no action if the sound is not from the current playlist
        if playlist != self._current_playlist:
            return

        # Determine if this is the last sound in the playlist
        if playlist.end_of_playlist and not playlist.repeat:
            return

        # Advance to the next sound in the playlist
        self.advance()

    def clear_context(self, context):
        """
        Stops the current playlist if it was played from the specified context.

        Args:
            context: The context to clear
        """
        self.log.debug("Clearing context %s", context)

        if self._current_playlist and self._current_playlist.context == context:
            self.stop()

        # Also need to check if there is a pending request to play another playlist with
        # the specified context.  If so, delete the pending request.
        if self._pending_request and "context" in self._pending_request.keywords and \
                self._pending_request.keywords["context"] == context:
            self._pending_request = None



# Tests the Image Asset and the Image widget


from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase


class TestImage(MpfMcTestCase, MpfSlideTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/assets_and_image'

    def get_config_file(self):
        return 'test_image.yaml'

    def test_image(self):
        self.mc.events.post('show_slide1')

        # This tests includes images that preload and that load on demand, so
        # give it enough time to for the on demand ones to load
        self.advance_time()

        # Make sure that all the images are showing.
        active_widget_names = [
            x.widget.image.name for x in self.mc.targets['default'].current_slide.widgets]

        for x in range(12):
            self.assertIn('image{}'.format(x+1), active_widget_names)

    def test_image_pools(self):
        self.mc.events.post("show_random_slide")
        self.advance_time(.1)
        self.assertSlideOnTop("random_image_test")
        self.mc.events.post("add_random_image")
        self.advance_time(.1)
        self.mc.events.post("remove_random_image")
        self.advance_time(.1)



"""Contains the parent class for Scriptlets used in MPF-MC."""

MYPY = False
if MYPY:    # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Scriptlet:

    """DEPRECATED scriptlet in MPF-MC."""

    def __init__(self, mc, name):
        self.mc = mc
        self.name = name
        self.on_load()

    def __repr__(self):  # pragma: no cover
        return '<Scriptlet.{}>'.format(self.name)

    def on_load(self):  # pragma: no cover
        """Automatically called when this Scriptlet loads. It's the intention
        that the Scriptlet writer will overwrite this method in the Scriptlet.
        """



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase


class TestVideo(MpfIntegrationTestCase):

    def get_machine_path(self):
        return 'integration/machine_files/video'

    def get_config_file(self):
        return 'config.yaml'

    def test_video_stops_on_slide_removal(self):
        # start mode1, its slide with video should come up
        self.post_event("start_mode1")
        self.advance_time_and_run()

        self.assertEqual(self.mc.targets['default'].current_slide.name, 'mode1_slide1')
        video_widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertEqual(video_widget.state, 'play')
        self.assertTrue(video_widget.video.loaded)

        # stop mode 1, should unload the slide
        self.post_event("stop_mode1")
        self.advance_time_and_run()
        self.assertEqual(self.mc.targets['default'].current_slide.name, 'default_blank')
        # check that the video is unloaded (because the widget will not be stopped)
        self.assertFalse(video_widget.video.loaded)

    def test_video_control_events_from_mpf(self):
        # start mode1, its slide with video should come up
        self.post_event("start_mode1")
        self.advance_time_and_run()

        self.assertEqual(self.mc.targets['default'].current_slide.name, 'mode1_slide1')
        video_widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertEqual(video_widget.state, 'play')
        self.assertTrue(video_widget.video.loaded)

        self.post_event("pause_event_from_mpf")
        self.advance_time_and_run(.1)
        self.assertEqual(video_widget.state, 'pause')

    def test_video_control_events_from_mpf_in_show(self):
        self.post_event("play_show_with_video")
        self.advance_time_and_run(.1)
        video_widget = self.mc.targets['default'].current_slide.widgets[0].widget
        self.assertEqual(video_widget.state, 'play')
        self.assertTrue(video_widget.video.loaded)

        self.post_event("pause_event_from_mpf2")
        self.advance_time_and_run(.1)
        self.assertEqual(video_widget.state, 'pause')



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestText(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/text'

    def get_config_file(self):
        return 'test_text.yaml'

    def get_widget(self, index=0):
        return self.mc.targets['default'].current_slide.widgets[index].widget

    def test_static_text(self):
        # Very basic test
        self.mc.events.post('static_text')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'TEST')

    def test_text_from_event_param1(self):
        # widget text is only from event param
        self.mc.events.post('text_from_event_param1', param1='HELLO')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'HELLO')

        # now make sure if we post the event again, the text updates
        self.mc.events.post('text_from_event_param1', param1='THIS')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'THIS')

        self.mc.events.post('text_from_event_param1', param1='IS')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'IS')

        self.mc.events.post('text_from_event_param1', param1='A')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'A')

        self.mc.events.post('text_from_event_param1', param1='NEW')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'NEW')

        self.mc.events.post('text_from_event_param1', param1='EVENT')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'EVENT')

    def test_text_from_event_param2(self):
        # widget text puts static text before param text
        self.mc.events.post('text_from_event_param2', param1='HELLO')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'HI HELLO')

    def test_text_from_event_param3(self):
        # widget text puts static text before and after param text
        self.mc.events.post('text_from_event_param3', param1='AND')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'MIX AND MATCH')

    def test_text_from_event_param4(self):
        # static and event text with no space between
        self.mc.events.post('text_from_event_param4', param1='SPACE')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'NOSPACE')

    def test_text_from_event_param5(self):
        #test event text that comes in as non-string
        self.mc.events.post('text_from_event_param5', param1=1)
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'NUMBER 1')

    def test_text_from_event_param6(self):
        # placeholder for event text for a param that doesn't exist
        self.mc.events.post('text_from_event_param6')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '(param1)')

    def test_text_from_event_param7(self):
        # test percent sign hard coded
        self.mc.events.post('text_from_event_param7')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '1)')

    def test_text_from_event_param8(self):
        # test parenthesis next to placeholder text
        self.mc.events.post('text_from_event_param8', param1=100)
        self.advance_time()

        self.assertEqual(self.get_widget().text, '(100)')

    def test_player_var1(self):
        # staight var, no player specified
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        self.mc.player.test_var = 1

        self.mc.events.post('text_with_player_var1')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '1')
        old_width = self.get_widget().width

        # update var, should update widget
        for x in range(99):
            self.mc.player.test_var += 1
            self.advance_time(.01)

        self.advance_time()
        self.assertEqual(self.get_widget().text, '100')
        self.assertGreater(self.get_widget().width, old_width)

    def test_player_var2(self):
        # 'player' specified
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        self.mc.player.test_var = 1

        self.mc.events.post('text_with_player_var2')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '1')

    def test_player_var3(self):
        # 'player1' specified
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        self.mc.player.test_var = 1

        self.mc.events.post('text_with_player_var3')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '1')

    def test_player_var4(self):
        # 'player2' specified with no player 2. Should be blank.
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        self.mc.player.test_var = 1

        self.mc.events.post('text_with_player_var4')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '')

        # Add player 2 & set the value. Widget should update
        self.mc.add_player(2)
        self.mc.player_list[1].test_var = 'Player 2 test variable'
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Player 2 test variable')

    def test_position_rounding(self):
        # staight var, no player specified
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        # set var, should update widget with even pixel width and not offset
        self.mc.player.test_var = 'its even'
        self.mc.events.post('text_with_player_var1')
        self.advance_time()

        self.get_widget()._round_anchor_styles = ('left', None)

        bounding_box = self.get_widget().canvas.children[-1]
        self.assertEqual(self.get_widget().text, 'its even')
        self.assertIn(bounding_box.size, ((346, 118), (343.0, 118.0)))
        self.assertEqual(bounding_box.pos, (200, 150))

        # update var, should update widget with an odd pixel width and offset DOWN
        self.mc.player.test_var = 'odd'
        self.advance_time()

        bounding_box = self.get_widget().canvas.children[-1]
        self.assertEqual(self.get_widget().text, 'odd')
        self.assertIn(bounding_box.size, ((171, 118), (169.0, 118.0)))
        self.assertEqual(bounding_box.pos, (199.5, 150))

        # update var, should update widget with an odd pixel width and offset UP
        self.get_widget()._round_anchor_styles = ('right', None)
        self.mc.player.test_var = 'also odd'
        self.advance_time()

        bounding_box = self.get_widget().canvas.children[-1]
        self.assertEqual(self.get_widget().text, 'also odd')
        self.assertIn(bounding_box.size, ((384, 118), (381.0, 118.0)))
        self.assertIn(bounding_box.pos, ((200.0, 150), (200.5, 150.0)))

        # update var, should update widget with an even pixel width and not offset
        self.mc.player.test_var = 'no round'
        self.advance_time()

        bounding_box = self.get_widget().canvas.children[-1]
        self.assertEqual(self.get_widget().text, 'no round')
        self.assertIn(bounding_box.size, ((396, 118), (394.0, 118.0)))
        self.assertEqual(bounding_box.pos, (200, 150))

    def test_current_player(self):
        # verifies that current player text update when current player changes
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)  # Player 1
        self.advance_time()
        self.mc.add_player(2)  # Player 2
        self.advance_time()
        self.mc.add_player(3)  # Player 3
        self.advance_time()

        # Test text: (test_var)

        # Player 1
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertEqual(0, self.mc.player.index)
        self.assertEqual(1, self.mc.player.number)

        self.mc.player.test_var = 'Player 1 test var'
        self.mc.events.post('text_with_player_var1')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'Player 1 test var')

        # Player 2
        self.mc.player_start_turn(2)
        self.advance_time()

        self.assertEqual(1, self.mc.player.index)
        self.assertEqual(2, self.mc.player.number)

        self.mc.player.test_var = 'Player 2 test var'
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'Player 2 test var')

        # Player 3
        self.mc.player_start_turn(3)
        self.advance_time()

        self.assertEqual(2, self.mc.player.index)
        self.assertEqual(3, self.mc.player.number)

        self.mc.player.test_var = 'Player 3 test var'
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'Player 3 test var')

        # Back to player 1, make sure it's still there
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'Player 1 test var')

        # Test text: (player|test_var)
        self.mc.events.post('text_with_player_var2')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Player 1 test var')

        self.mc.player_start_turn(2)
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Player 2 test var')

        self.mc.player_start_turn(3)
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Player 3 test var')

    def test_mix_player_var_and_event_param(self):
        self.mc.game_start()
        self.advance_time()
        self.mc.add_player(1)
        self.advance_time()
        self.mc.player_start_turn(1)
        self.advance_time()

        self.assertTrue(self.mc.player)

        self.mc.player.player_var = 'PLAYER VAR'

        self.mc.events.post('text_with_player_var_and_event',
                            test_param="EVENT PARAM")
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'PLAYER VAR EVENT PARAM')

        self.mc.player.player_var = 'NEW PLAYER VAR'
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'NEW PLAYER VAR EVENT PARAM')

        self.mc.events.post('text_with_player_var_and_event',
                            test_param="NEW EVENT PARAM")
        self.advance_time()
        self.assertEqual(self.get_widget().text,
                         'NEW PLAYER VAR NEW EVENT PARAM')

    def test_number_grouping(self):
        self.mc.events.post('number_grouping')
        self.advance_time()

        # should be 00 even though text is 0
        self.assertEqual(self.get_widget().text, '00')
        self.advance_time()

        self.get_widget().update_text('2000000')
        self.assertEqual(self.get_widget().text, '2,000,000')
        self.advance_time()

    def test_text_casing(self):
        self.mc.events.post('text_nocase')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'sAmPlE tExT caSiNg')

        self.mc.events.post('text_lower')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'sample text casing')

        self.mc.events.post('text_upper')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'SAMPLE TEXT CASING')

        self.mc.events.post('text_title')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Sample Text Casing')

        self.mc.events.post('text_capitalize')
        self.advance_time()
        self.assertEqual(self.get_widget().text, 'Sample text casing')

    def test_text_string1(self):
        # simple text string in machine config
        self.mc.events.post('text_string1')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'HELLO')

    def test_text_string2(self):
        # two text strings in machine config
        self.mc.events.post('text_string2')
        self.advance_time()

        self.assertEqual(self.get_widget().text, 'HELLO PLAYER')

    def test_text_string3(self):
        # text string not found
        self.mc.events.post('text_string3')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '$money')

    def test_text_string4(self):
        # text string found with extra dollar sign in text
        self.mc.events.post('text_string4')
        self.advance_time()

        self.assertEqual(self.get_widget().text, '$100')

    def test_custom_fonts(self):
        self.mc.events.post('mpfmc_font')
        self.advance_time()
        self.assertEqual('pixelmix', self.get_widget().font_name)
        self.mc.events.post('machine_font')
        self.advance_time()
        self.assertEqual('big_noodle_titling', self.get_widget().font_name)

    def test_baseline(self):
        self.mc.events.post('baseline')
        self.advance_time()

        # Baseline anchored widgets should all have the same local coordinates
        self.assertEqual(self.get_widget(0).pos[1], 100)
        self.assertEqual(self.get_widget(1).pos[1], 100)
        self.assertEqual(self.get_widget(2).pos[1], 100)
        self.assertEqual(self.get_widget(3).pos[1], 100)

        # Now convert the widget coordinates to parent coordinates for comparison
        widget0 = self.get_widget(0).parent.to_parent(0, self.get_widget(0).pos[1])
        widget1 = self.get_widget(1).parent.to_parent(0, self.get_widget(1).pos[1])
        widget2 = self.get_widget(2).parent.to_parent(0, self.get_widget(2).pos[1])
        widget3 = self.get_widget(3).parent.to_parent(0, self.get_widget(3).pos[1])

        # Baseline anchored widgets should be 4px lower
        self.assertEqual(widget0[1], 100)
        self.assertEqual(widget1[1], 96)
        self.assertEqual(widget2[1], 100)
        self.assertEqual(widget3[1], 96)

    def test_line_break(self):
        """Tests line break in text (multiple lines)"""
        self.mc.events.post('text_line_break')
        self.advance_time()
        self.assertGreater(self.get_widget().height, 40)

    def test_no_multiline(self):
        """Tests poorly formatted YAML line break in text"""
        self.mc.events.post('text_bad_line_break')
        self.advance_time()
        self.assertLess(self.get_widget().height, 30)



from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDmd(MpfMcTestCase, MpfSlideTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/dmd'

    def get_config_file(self):
        return 'test_dmd.yaml'

    def test_dmd(self):

        self.assertIn('dmd', self.mc.targets)

        self.mc.events.post('container_slide')
        self.mc.events.post('dmd_slide')
        self.advance_time(2)

    def test_positioning_named_widgets_on_dmd(self):

        self.assertIn('dmd', self.mc.targets)

        self.mc.events.post('container_slide')
        self.mc.events.post('position_widget_left')
        self.advance_time(.1)
        self.mc.events.post('position_widget_right')
        self.advance_time(.1)
        self.mc.events.post('position_widget_top')
        self.advance_time(.1)
        self.mc.events.post('position_widget_bottom')
        self.advance_time(.1)

        left = self.mc.displays['dmd'].current_slide.widgets[0].widget
        right = self.mc.displays['dmd'].current_slide.widgets[1].widget
        top = self.mc.displays['dmd'].current_slide.widgets[2].widget
        bottom = self.mc.displays['dmd'].current_slide.widgets[3].widget

        self.assertEqual(left.text, 'Left Widget')
        self.assertEqual(right.text, 'Right Widget')
        self.assertEqual(top.text, 'Top Widget')
        self.assertEqual(bottom.text, 'Bottom Widget')

        self.assertGreater(right.x, bottom.x)
        self.assertLess(left.x, bottom.x)
        self.assertGreater(top.y, left.y)
        self.assertLess(bottom.y, left.y)

        # trigger the same widget again
        self.mc.events.post('position_widget_left')
        self.advance_time(.1)
        self.assertEqual(left.text, 'Left Widget')
        self.assertEqual(4, len(self.mc.displays['dmd'].current_slide.widgets))

    def test_gamma_test_slide(self):
        self.mc.events.post('show_gamma_test')
        self.advance_time(.1)
        self.assertSlideOnTop("dmd_gamma_test")
